# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Zachary's karate club example.

Here we train a graph neural network to process Zachary's karate club.
https://en.wikipedia.org/wiki/Zachary%27s_karate_club

Zachary's karate club is used in the literature as an example of a social graph.
Here we use a graphnet to optimize the assignments of the students in the
karate club to two distinct karate instructors (Mr. Hi and John A).
"""

import logging

from absl import app
import haiku as hk
import jax
import jax.numpy as jnp
import jraph
import optax


def get_zacharys_karate_club() -> jraph.GraphsTuple:
  """Returns GraphsTuple representing Zachary's karate club."""
  social_graph = [
      (1, 0), (2, 0), (2, 1), (3, 0), (3, 1), (3, 2),
      (4, 0), (5, 0), (6, 0), (6, 4), (6, 5), (7, 0), (7, 1),
      (7, 2), (7, 3), (8, 0), (8, 2), (9, 2), (10, 0), (10, 4),
      (10, 5), (11, 0), (12, 0), (12, 3), (13, 0), (13, 1), (13, 2),
      (13, 3), (16, 5), (16, 6), (17, 0), (17, 1), (19, 0), (19, 1),
      (21, 0), (21, 1), (25, 23), (25, 24), (27, 2), (27, 23),
      (27, 24), (28, 2), (29, 23), (29, 26), (30, 1), (30, 8),
      (31, 0), (31, 24), (31, 25), (31, 28), (32, 2), (32, 8),
      (32, 14), (32, 15), (32, 18), (32, 20), (32, 22), (32, 23),
      (32, 29), (32, 30), (32, 31), (33, 8), (33, 9), (33, 13),
      (33, 14), (33, 15), (33, 18), (33, 19), (33, 20), (33, 22),
      (33, 23), (33, 26), (33, 27), (33, 28), (33, 29), (33, 30),
      (33, 31), (33, 32)]
  # Add reverse edges.
  social_graph += [(edge[1], edge[0]) for edge in social_graph]
  n_club_members = 34

  return jraph.GraphsTuple(
      n_node=jnp.asarray([n_club_members]),
      n_edge=jnp.asarray([len(social_graph)]),
      # One-hot encoding for nodes.
      nodes=jnp.eye(n_club_members),
      # No edge features.
      edges=None,
      globals=None,
      senders=jnp.asarray([edge[0] for edge in social_graph]),
      receivers=jnp.asarray([edge[1] for edge in social_graph]))


def get_ground_truth_assignments_for_zacharys_karate_club() -> jnp.ndarray:
  """Returns ground truth assignments for Zachary's karate club."""
  return jnp.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1,
                    0, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])


def network_definition(graph: jraph.GraphsTuple) -> jraph.ArrayTree:
  """Implements the GCN from Kipf et al https://arxiv.org/pdf/1609.02907.pdf.

     A' = D^{-0.5} A D^{-0.5}
     Z = f(X, A') = A' relu(A' X W_0) W_1

  Args:
    graph: GraphsTuple the network processes.

  Returns:
    processed nodes.
  """
  gn = jraph.GraphConvolution(
      update_node_fn=hk.Linear(5, with_bias=False),
      add_self_edges=True)
  graph = gn(graph)
  graph = graph._replace(nodes=jax.nn.relu(graph.nodes))
  gn = jraph.GraphConvolution(
      update_node_fn=hk.Linear(2, with_bias=False))
  graph = gn(graph)
  return graph.nodes


def optimize_club(num_steps: int):
  """Solves the karte club problem by optimizing the assignments of students."""
  network = hk.without_apply_rng(hk.transform(network_definition))
  zacharys_karate_club = get_zacharys_karate_club()
  labels = get_ground_truth_assignments_for_zacharys_karate_club()
  params = network.init(jax.random.PRNGKey(42), zacharys_karate_club)

  @jax.jit
  def prediction_loss(params):
    decoded_nodes = network.apply(params, zacharys_karate_club)
    # We interpret the decoded nodes as a pair of logits for each node.
    log_prob = jax.nn.log_softmax(decoded_nodes)
    # The only two assignments we know a-priori are those of Mr. Hi (Node 0)
    # and John A (Node 33).
    return -(log_prob[0, 0] + log_prob[33, 1])

  opt_init, opt_update = optax.adam(1e-2)
  opt_state = opt_init(params)

  @jax.jit
  def update(params, opt_state):
    g = jax.grad(prediction_loss)(params)
    updates, opt_state = opt_update(g, opt_state)
    return optax.apply_updates(params, updates), opt_state

  @jax.jit
  def accuracy(params):
    decoded_nodes = network.apply(params, zacharys_karate_club)
    return jnp.mean(jnp.argmax(decoded_nodes, axis=1) == labels)

  for step in range(num_steps):
    logging.info("step %r accuracy %r", step, accuracy(params).item())
    params, opt_state = update(params, opt_state)


def main(_):
  optimize_club(num_steps=30)


if __name__ == "__main__":
  app.run(main)



# Copyright 2020 DeepMind Technologies Limited.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sharded (Data Parallel) Graph Nets."""

import functools
from typing import Callable, List, NamedTuple, Optional
import jax
import jax.numpy as jnp
import jax.tree_util as tree
import jraph
from jraph._src import graph as gn_graph
from jraph._src import utils
import numpy as np


class ShardedEdgesGraphsTuple(NamedTuple):
  """A `GraphsTuple` for use with `ShardedEdgesGraphNetwork`.

  NOTES:
  - A ShardedEdgesGraphNetwork is for use with `jax.pmap`. As such, it will have
    a leading axis of size `num_devices` on the host, but no such axis on
    device. Non-sharded data is replicated on each device. To achieve this with
    `jax.pmap` you can broadcast non-sharded data to have leading axis
    'num_devices' or use the 'in_axes' parameter, which will indicate which
    attributes should be replicated and which should not. Current helper methods
    use the first approach.
  - It is recommended that you constructed ShardedEdgesGraphsTuples with
    `graphs_tuple_to_broadcasted_sharded_grahs_tuple`.


  The values of `nodes`, `device_edges` and `globals` can be gn_graph.ArrayTree
  - nests of features with `jax` compatible values. For example, `nodes` in a
  graph may have more than one type of attribute.

  However, the ShardedEdgesGraphsTuple typically takes the following form for a
  batch of `n` graphs:

  - n_node: The number of nodes per graph. It is a vector of integers with shape
    `[n_graphs]`, such that `graph.n_node[i]` is the number of nodes in the i-th
    graph.

  - n_edge: The number of edges per graph. It is a vector of integers with shape
    `[n_graphs]`, such that `graph.n_edge[i]` is the number of edges in the i-th
    graph.

  - nodes: The nodes features. It is either `None` (the graph has no node
    features), or a vector of shape `[n_nodes] + node_shape`, where
    `n_nodes = sum(graph.n_node)` is the total number of nodes in the batch of
    graphs, and `node_shape` represents the shape of the features of each node.
    The relative index of a node from the batched version can be recovered from
    the `graph.n_node` property. For instance, the second node of the third
    graph will have its features in the
    `1 + graph.n_node[0] + graph.n_node[1]`-th slot of graph.nodes.
    Observe that having a `None` value for this field does not mean that the
    graphs have no nodes, only that they do not have node features.

  - receivers: The indices of the receiver nodes, for each edge. It is either
    `None` (if the graph has no edges), or a vector of integers of shape
    `[n_edges]`, such that `graph.receivers[i]` is the index of the node
    receiving from the i-th edge.

    Observe that the index is absolute (in other words, cumulative), i.e.
    `graphs.receivers` take value in `[0, n_nodes]`. For instance, an edge
    connecting the vertices with relative indices 2 and 3 in the second graph of
    the batch would have a `receivers` value of `3 + graph.n_node[0]`.
    If `graphs.receivers` is `None`, then `graphs.edges` and `graphs.senders`
    should also be `None`.

  - senders: The indices of the sender nodes, for each edge. It is either
    `None` (if the graph has no edges), or a vector of integers of shape
    `[n_edges]`, such that `graph.senders[i]` is the index of the node
    sending from the i-th edge.

    Observe that the index is absolute, i.e. `graphs.senders` take value in
    `[0, n_nodes]`. For instance, an edge connecting the vertices with relative
    indices 1 and 3 in the third graph of the batch would have a `senders` value
    of `1 + graph.n_node[0] + graph.n_node[1]`.

    If `graphs.senders` is `None`, then `graphs.edges` and `graphs.receivers`
    should also be `None`.

  - globals: The global features of the graph. It is either `None` (the graph
    has no global features), or a vector of shape `[n_graphs] + global_shape`
    representing graph level features.

  The ShardedEdgesGraphsTuple also contains device-local attributes that are
  used for data parallel computation. On the host, each of these attributes will
  have an additional leading axis of shape `num_devices` for use with
  `jax.pmap`, but this is ommited in the following documentation.

  - device_edges: The subset of the edge features that are on the device.
      It is either `None` (the graph has no edge features), or a vector of
      shape `[num_edges / num_devices] + edge_shape`

      Observe that having a `None` value for this field does not mean that the
      graph has no edges, only that they do not have edge features.

  - device_senders: The sender indices of edges on device. This is of length
      num_edges / num_devices.

  - device_receivers: The receiver indices of edge on device. This is of length
      num_edges / num_devices.

  - device_n_edge: The graph partitions of the edges on device. For example,
      say that there are 2 graphs in the original graphs tuple, with n_edge
      [1, 11], which has been split over 3 devices. The `device_n_edge`s would
      be [[1, 3], [4, 0], [4, 0]]. `0` valued entries that are padding values or
      graphs with zero edges are not distinguished. Since these attributes are
      used only for `repeat` purposes, the difference makes no difference to
      the implementation.

  - device_graph_idx: The indices of the graphs on device. For example, say
      that there are 5 graphs in the original graphs tuple, and these has been
      split over 3 devices, the device_graphs_idxs could be
      [[0, 1, 2], [2, 3, 0], [3, 4, 0]]. In this splitting, the third graph
      is split over 2 devices. If a `0` is the first in `device_graph_idx` then
      that indicates the first graph, otherwise it indicates a padding value.
  """
  nodes: gn_graph.ArrayTree
  device_edges: gn_graph.ArrayTree
  device_receivers: jnp.ndarray  # with integer dtype
  device_senders: jnp.ndarray  # with integer dtype
  receivers: jnp.ndarray  # with integer dtype
  senders: jnp.ndarray  # with integer dtype
  globals: gn_graph.ArrayTree
  device_n_edge: jnp.ndarray  # with integer dtype
  n_node: jnp.ndarray  # with integer dtype
  n_edge: jnp.ndarray  # with integer dtype
  device_graph_idx: jnp.ndarray  # with integer dtype


def graphs_tuple_to_broadcasted_sharded_graphs_tuple(
    graphs_tuple: jraph.GraphsTuple,
    num_shards: int) -> ShardedEdgesGraphsTuple:
  """Converts a `GraphsTuple` to a `ShardedEdgesGraphsTuple` to use with `pmap`.

  For a given number of shards this will compute device-local edge and graph
  attributes, and add a batch axis of size num_shards. You can then use
  `ShardedEdgesGraphNetwork` with `jax.pmap`.

  Args:
    graphs_tuple: The `GraphsTuple` to be converted to a sharded `GraphsTuple`.
    num_shards: The number of devices to shard over.

  Returns:
    A ShardedEdgesGraphsTuple over the number of shards.
  """
  # Note: this is not jittable, so to prevent using a device by accident,
  # this is all happening in numpy.
  nodes, edges, receivers, senders, globals_, n_node, n_edge = graphs_tuple
  if np.sum(n_edge) % num_shards != 0:
    raise ValueError(('The number of edges in a `graph.GraphsTuple` must be '
                      'divisible by the number of devices per replica.'))
  if np.sum(np.array(n_edge)) == 0:
    raise ValueError('The input `Graphstuple` must have edges.')
  # Broadcast replicated features to have a `num_shards` leading axis.
  # pylint: disable=g-long-lambda
  broadcast = lambda x: np.broadcast_to(x[None, :], (num_shards,) + x.shape)
  # pylint: enable=g-long-lambda

  # `edges` will be straightforwardly sharded, with 1/num_shards of
  # the edges on each device.
  def shard_edges(edge_features):
    return np.reshape(edge_features, (num_shards, -1) + edge_features.shape[1:])

  edges = jax.tree_map(shard_edges, edges)
  # Our sharded strategy is by edges - which means we need a device local
  # n_edge, senders and receivers to do global aggregations.

  # Senders and receivers are easy - 1/num_shards per device.
  device_senders = shard_edges(senders)
  device_receivers = shard_edges(receivers)

  # n_edge is a bit more difficult. Let's say we have a graphs tuple with
  # n_edge [2, 8], and we want to distribute this on two devices. Then
  # we will have sharded the edges to [5, 5], so the n_edge per device will be
  # [2,3], and [5]. Since we need to have each of the n_edge the same shape,
  # we will need to pad this to [5,0]. This is a bit dangerous, as the zero
  # here has a different meaning to a graph with zero edges, but we need the
  # zero for the global broadcasting to be correct for aggregation. Since
  # this will only be used in the first instance for global broadcasting on
  # device I think this is ok, but ideally we'd have a more elegant solution.
  # TODO(jonathangodwin): think of a more elegant solution.
  edges_per_device = np.sum(n_edge) // num_shards
  edges_in_current_split = 0
  completed_splits = []
  current_split = {'n_edge': [], 'device_graph_idx': []}
  for device_graph_idx, x in enumerate(n_edge):
    new_edges_in_current_split = edges_in_current_split + x
    if new_edges_in_current_split > edges_per_device:
      # A single graph may be spread across multiple replicas, so here we
      # iteratively create new splits until the graph is exhausted.

      # How many edges we are trying to allocate.
      carry = x
      # How much room there is in the current split for new edges.
      space_in_current_split = edges_per_device - edges_in_current_split
      while carry > 0:
        if carry >= space_in_current_split:
          # We've encountered a situation where we need to split a graph across
          # >= 2 devices. We compute the number we will carry to the next split,
          # and add a full split.
          carry = carry - space_in_current_split
          # Add the left edges to the current split, and complete the split
          # by adding it to completed_splits.
          current_split['n_edge'].append(space_in_current_split)
          current_split['device_graph_idx'].append(device_graph_idx)
          completed_splits.append(current_split)
          # reset the split
          current_split = {'n_edge': [], 'device_graph_idx': []}

          space_in_current_split = edges_per_device
          edges_in_current_split = 0
        else:
          current_split = {
              'n_edge': [carry],
              'device_graph_idx': [device_graph_idx]
          }
          edges_in_current_split = carry
          carry = 0
          # Since the total number of edges must be divisible by the number
          # of devices, this code  path can only be executed for an intermediate
          # graph, thus it is not a complete split and we never need to add it
          # to `completed splits`.
    else:
      # Add the edges and globals to the current split.
      current_split['n_edge'].append(x)
      current_split['device_graph_idx'].append(device_graph_idx)
      # If we've reached the end of a split, complete it and start a new one.
      if new_edges_in_current_split == edges_per_device:
        completed_splits.append(current_split)
        current_split = {'n_edge': [], 'device_graph_idx': []}
        edges_in_current_split = 0
      else:
        edges_in_current_split = new_edges_in_current_split

  # Flatten list of dicts to dict of lists.
  completed_splits = {
      k: [d[k] for d in completed_splits] for k in completed_splits[0]
  }
  pad_split_to = max([len(x) for x in completed_splits['n_edge']])
  pad = lambda x: np.pad(x, (0, pad_split_to - len(x)), mode='constant')
  device_n_edge = np.array([pad(x) for x in completed_splits['n_edge']])
  device_graph_idx = np.array(
      [pad(x) for x in completed_splits['device_graph_idx']])
  return ShardedEdgesGraphsTuple(
      nodes=jax.tree_map(broadcast, nodes),
      device_edges=edges,
      device_receivers=device_receivers,
      device_senders=device_senders,
      receivers=broadcast(receivers),
      senders=broadcast(senders),
      device_graph_idx=device_graph_idx,
      globals=jax.tree_map(broadcast, globals_),
      n_node=broadcast(n_node),
      n_edge=broadcast(n_edge),
      device_n_edge=device_n_edge)


def broadcasted_sharded_graphs_tuple_to_graphs_tuple(sharded_graphs_tuple):
  """Converts a broadcasted ShardedGraphsTuple to a GraphsTuple."""
  # We index the first element of replicated arrays, since they have been
  # repeated. For edges, we reshape to recover all of the edge features.
  unbroadcast = lambda y: tree.tree_map(lambda x: x[0], y)
  unshard = lambda x: jnp.reshape(x, (x.shape[0] * x.shape[1],) + x.shape[2:])
  # TODO(jonathangodwin): check senders and receivers are consistent.
  return jraph.GraphsTuple(
      nodes=unbroadcast(sharded_graphs_tuple.nodes),
      edges=tree.tree_map(unshard, sharded_graphs_tuple.device_edges),
      n_node=sharded_graphs_tuple.n_node[0],
      n_edge=sharded_graphs_tuple.n_edge[0],
      globals=unbroadcast(sharded_graphs_tuple.globals),
      senders=sharded_graphs_tuple.senders[0],
      receivers=sharded_graphs_tuple.receivers[0])


def sharded_segment_sum(data, indices, num_segments, axis_index_groups):
  """Segment sum over data on multiple devices."""
  device_segment_sum = utils.segment_sum(data, indices, num_segments)
  return jax.lax.psum(
      device_segment_sum, axis_name='i', axis_index_groups=axis_index_groups)


ShardedEdgeFeatures = gn_graph.ArrayTree
AggregateShardedEdgesToGlobalsFn = Callable[
    [ShardedEdgeFeatures, jnp.ndarray, int, jnp.ndarray], gn_graph.ArrayTree]
AggregateShardedEdgesToNodesFn = Callable[
    [gn_graph.ArrayTree, jnp.ndarray, int, List[List[int]]], jraph.NodeFeatures]


# pylint: disable=invalid-name
def ShardedEdgesGraphNetwork(
    update_edge_fn: Optional[jraph.GNUpdateEdgeFn],
    update_node_fn: Optional[jraph.GNUpdateNodeFn],
    update_global_fn: Optional[jraph.GNUpdateGlobalFn] = None,
    aggregate_edges_for_nodes_fn:
    AggregateShardedEdgesToNodesFn = sharded_segment_sum,
    aggregate_nodes_for_globals_fn: jraph.AggregateNodesToGlobalsFn = jax.ops
    .segment_sum,
    aggregate_edges_for_globals_fn:
    AggregateShardedEdgesToGlobalsFn = sharded_segment_sum,
    attention_logit_fn: Optional[jraph.AttentionLogitFn] = None,
    attention_reduce_fn: Optional[jraph.AttentionReduceFn] = None,
    num_shards: int = 1):
  """Returns a method that applies a GraphNetwork on a sharded GraphsTuple.

  This GraphNetwork is sharded over `edges`, all other features are assumed
  to be replicated on device.
  There are two clear use cases for a ShardedEdgesGraphNetwork. The first is
  where a single graph can't fit on device. The second is when you are compute
  bound on the edge feature calculation, and you'd like to speed up
  training/inference by distributing the compute across devices.

  Example usage:

    ```
    gn = jax.pmap(ShardedEdgesGraphNetwork(update_edge_function,
    update_node_function, **kwargs), axis_name='i')
    # Conduct multiple rounds of message passing with the same parameters:
    for _ in range(num_message_passing_steps):
      sharded_graph = gn(sharded_graph)
    ```

  Args:
    update_edge_fn: function used to update the edges or None to deactivate edge
      updates.
    update_node_fn: function used to update the nodes or None to deactivate node
      updates.
    update_global_fn: function used to update the globals or None to deactivate
      globals updates.
    aggregate_edges_for_nodes_fn: function used to aggregate messages to each
      nodes. This must support cross-device aggregations.
    aggregate_nodes_for_globals_fn: function used to aggregate the nodes for the
      globals.
    aggregate_edges_for_globals_fn: function used to aggregate the edges for the
      globals. This must support cross-device aggregations.
    attention_logit_fn: function used to calculate the attention weights or None
      to deactivate attention mechanism.
    attention_reduce_fn: function used to apply weights to the edge features or
      None if attention mechanism is not active.
    num_shards: how many devices per replica for sharding.

  Returns:
    A method that applies the configured GraphNetwork.
  """
  not_both_supplied = lambda x, y: (x != y) and ((x is None) or (y is None))
  if not_both_supplied(attention_reduce_fn, attention_logit_fn):
    raise ValueError(('attention_logit_fn and attention_reduce_fn must both be'
                      ' supplied.'))

  devices = jax.devices()
  num_devices = len(devices)
  assert num_devices % num_shards == 0
  num_replicas = num_devices // num_shards
  # The IDs within a replica.
  replica_ids = list(range(num_devices))
  # How the devices are grouped per replica.
  axis_groups = [
      replica_ids[i * num_shards:(i + 1) * num_shards]
      for i in range(num_replicas)
  ]

  def _ApplyGraphNet(graph: ShardedEdgesGraphsTuple) -> ShardedEdgesGraphsTuple:
    """Applies a configured GraphNetwork to a sharded graph.

    This implementation follows Algorithm 1 in https://arxiv.org/abs/1806.01261

    There is one difference. For the nodes update the class aggregates over the
    sender edges and receiver edges separately. This is a bit more general
    the algorithm described in the paper. The original behaviour can be
    recovered by using only the receiver edge aggregations for the update.

    In addition this implementation supports softmax attention over incoming
    edge features.


    Many popular Graph Neural Networks can be implemented as special cases of
    GraphNets, for more information please see the paper.

    Args:
      graph: a `GraphsTuple` containing the graph.

    Returns:
      Updated `GraphsTuple`.
    """
    # pylint: disable=g-long-lambda
    nodes, device_edges, device_receivers, device_senders, receivers, senders, globals_, device_n_edge, n_node, n_edge, device_graph_idx = graph
    # Equivalent to jnp.sum(n_node), but jittable.
    sum_n_node = tree.tree_leaves(nodes)[0].shape[0]
    sum_device_n_edge = device_senders.shape[0]
    if not tree.tree_all(
        tree.tree_map(lambda n: n.shape[0] == sum_n_node, nodes)):
      raise ValueError(
          'All node arrays in nest must contain the same number of nodes.')

    sent_attributes = tree.tree_map(lambda n: n[device_senders], nodes)
    received_attributes = tree.tree_map(lambda n: n[device_receivers], nodes)
    # Here we scatter the global features to the corresponding edges,
    # giving us tensors of shape [num_edges, global_feat].
    global_edge_attributes = tree.tree_map(
        lambda g: jnp.repeat(
            g[device_graph_idx], device_n_edge, axis=0,
            total_repeat_length=sum_device_n_edge),
        globals_)

    if update_edge_fn:
      device_edges = update_edge_fn(device_edges, sent_attributes,
                                    received_attributes, global_edge_attributes)

    if attention_logit_fn:
      logits = attention_logit_fn(device_edges, sent_attributes,
                                  received_attributes, global_edge_attributes)
      tree_calculate_weights = functools.partial(
          utils.segment_softmax, segment_ids=receivers, num_segments=sum_n_node)
      weights = tree.tree_map(tree_calculate_weights, logits)
      device_edges = attention_reduce_fn(device_edges, weights)

    if update_node_fn:
      # Aggregations over nodes are assumed to take place over devices
      # specified by the axis_groups (e.g. with sharded_segment_sum).
      sent_attributes = tree.tree_map(
          lambda e: aggregate_edges_for_nodes_fn(e, device_senders, sum_n_node,
                                                 axis_groups), device_edges)
      received_attributes = tree.tree_map(
          lambda e: aggregate_edges_for_nodes_fn(
              e, device_receivers, sum_n_node, axis_groups), device_edges)
      # Here we scatter the global features to the corresponding nodes,
      # giving us tensors of shape [num_nodes, global_feat].
      global_attributes = tree.tree_map(
          lambda g: jnp.repeat(
              g, n_node, axis=0, total_repeat_length=sum_n_node), globals_)
      nodes = update_node_fn(nodes, sent_attributes, received_attributes,
                             global_attributes)

    if update_global_fn:
      n_graph = n_node.shape[0]
      graph_idx = jnp.arange(n_graph)
      # To aggregate nodes and edges from each graph to global features,
      # we first construct tensors that map the node to the corresponding graph.
      # For example, if you have `n_node=[1,2]`, we construct the tensor
      # [0, 1, 1]. We then do the same for edges.
      node_gr_idx = jnp.repeat(
          graph_idx, n_node, axis=0, total_repeat_length=sum_n_node)
      edge_gr_idx = jnp.repeat(
          device_graph_idx,
          device_n_edge,
          axis=0,
          total_repeat_length=sum_device_n_edge)
      # We use the aggregation function to pool the nodes/edges per graph.
      node_attributes = tree.tree_map(
          lambda n: aggregate_nodes_for_globals_fn(n, node_gr_idx, n_graph),
          nodes)
      edge_attribtutes = tree.tree_map(
          lambda e: aggregate_edges_for_globals_fn(e, edge_gr_idx, n_graph,
                                                   axis_groups), device_edges)
      # These pooled nodes are the inputs to the global update fn.
      globals_ = update_global_fn(node_attributes, edge_attribtutes, globals_)
    # pylint: enable=g-long-lambda
    return ShardedEdgesGraphsTuple(
        nodes=nodes,
        device_edges=device_edges,
        device_senders=device_senders,
        device_receivers=device_receivers,
        receivers=receivers,
        senders=senders,
        device_graph_idx=device_graph_idx,
        globals=globals_,
        n_node=n_node,
        n_edge=n_edge,
        device_n_edge=device_n_edge)

  return _ApplyGraphNet
# pylint: enable=invalid-name



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Higgs Boson Detection Example.

One of the decay-channels of the Higgs Boson is Higgs to two photons.
The two photons must have a combined invariant mass 125 GeV.
In this example we use a relational vector to detect if a Higgs Boson
is present in a set of photons.

There are two situations:
a) Higgs: Two photons with an invariant mass of 125 GeV + an arbitrary number of
   uncorrelated photons.
b) No Higgs: Just an arbitrary number of uncorrelation photons.
"""

import collections
import logging
import random

from absl import app
import haiku as hk
import jax
import jax.numpy as jnp
import jraph
import numpy as np
import optax
import scipy.stats


Problem = collections.namedtuple("Problem", ("graph", "labels"))


def get_random_rotation_matrix():
  rotation = np.eye(4)
  rotation[1:, 1:] = scipy.stats.ortho_group.rvs(3)
  return rotation


def get_random_boost_matrix():
  eta = np.random.uniform(-1, 1)
  boost = np.eye(4)
  boost[:2, :2] = np.array([[np.cosh(eta), -np.sinh(eta)],
                            [-np.sinh(eta), np.cosh(eta)]])
  rotation = get_random_rotation_matrix()
  return rotation.T @ boost @ rotation


def get_random_higgs_photons():
  higgs = 125.18
  boost = get_random_boost_matrix()
  rotation = get_random_rotation_matrix()
  photon1 = boost @ rotation @ np.array([higgs / 2, higgs / 2, 0, 0])
  photon2 = boost @ rotation @ np.array([higgs / 2, -higgs / 2, 0, 0])
  return photon1, photon2


def get_random_background_photon():
  boost = get_random_boost_matrix()
  rotation = get_random_rotation_matrix()
  energy = np.random.uniform(20, 120)
  return boost @ rotation @ np.array([energy, energy, 0, 0])


def get_higgs_problem(min_n_photons: int, max_n_photons: int) -> Problem:
  """Creates fully connected graph containing the detected photons.

  Args:
    min_n_photons: minimum number of photons in the detector.
    max_n_photons: maximum number of photons in the detector.

  Returns:
    graph, one-hot label whether a higgs was present or not.
  """
  assert min_n_photons >= 2, "Number of photons must be at least 2."
  n_photons = random.randint(min_n_photons, max_n_photons)
  photons = np.stack([get_random_background_photon() for _ in range(n_photons)])

  # Add a higgs
  if random.random() > 0.5:
    label = np.eye(2)[0]
    photons[:2] = np.stack(get_random_higgs_photons())
  else:
    label = np.eye(2)[1]

  # The graph is fully connected.
  senders = np.repeat(np.arange(n_photons), n_photons)
  receivers = np.tile(np.arange(n_photons), n_photons)
  graph = jraph.GraphsTuple(
      n_node=np.asarray([n_photons]),
      n_edge=np.asarray([len(senders)]),
      nodes=photons,
      edges=None,
      globals=None,
      senders=senders,
      receivers=receivers)

  # In order to jit compile our code, we have to pad the nodes and edges of
  # the GraphsTuple to a static shape.
  graph = jraph.pad_with_graphs(graph, max_n_photons + 1,
                                max_n_photons * max_n_photons)

  return Problem(graph=graph, labels=label)


def network_definition(
    graph: jraph.GraphsTuple) -> jraph.ArrayTree:
  """Defines a graph neural network.

  Args:
    graph: Graphstuple the network processes.

  Returns:
    globals.
  """

  @jax.vmap
  @jraph.concatenated_args
  def update_edge_fn(features):
    return hk.nets.MLP([30, 30, 30])(features)

  # The correct solution for the edge update function is the invariant mass
  # of the photon pair.
  # The simple MLP we use here seems to fail to find the correct solution.
  # You can ensure that the example works in principle by replacing the
  # update_edge_fn below with the following analytical solution.
  @jax.vmap
  def unused_update_edge_fn_solution(s, r):
    """Calculates invariant mass of photon pair and compares to Higgs mass."""
    t = (s + r)**2
    return jnp.array(jnp.abs(t[0] - t[1] - t[2] - t[3] - 125.18**2) < 1,
                     dtype=jnp.float32)[None]

  gn = jraph.RelationNetwork(
      update_edge_fn=update_edge_fn,
      update_global_fn=hk.nets.MLP([2]),
      aggregate_edges_for_globals_fn=jraph.segment_sum,
      )
  graph = gn(graph)

  return graph.globals


def train(num_steps: int):
  """Trains a graph neural network on an electronic voting problem."""
  train_dataset = (2, 15)
  test_dataset = (16, 20)
  random.seed(42)

  network = hk.without_apply_rng(hk.transform(network_definition))
  problem = get_higgs_problem(*train_dataset)
  params = network.init(jax.random.PRNGKey(42), problem.graph)

  @jax.jit
  def prediction_loss(params, problem):
    globals_ = network.apply(params, problem.graph)
    # We interpret the globals as logits for the detection.
    # Only the first graph is real, the second graph is for padding.
    log_prob = jax.nn.log_softmax(globals_[0]) * problem.labels
    return -jnp.sum(log_prob)

  @jax.jit
  def accuracy_loss(params, problem):
    globals_ = network.apply(params, problem.graph)
    # We interpret the globals as logits for the detection.
    # Only the first graph is real, the second graph is for padding.
    equal = jnp.argmax(globals_[0]) == jnp.argmax(problem.labels)
    return equal.astype(np.int32)

  opt_init, opt_update = optax.adam(2e-4)
  opt_state = opt_init(params)

  @jax.jit
  def update(params, opt_state, problem):
    g = jax.grad(prediction_loss)(params, problem)
    updates, opt_state = opt_update(g, opt_state)
    return optax.apply_updates(params, updates), opt_state

  for step in range(num_steps):
    problem = get_higgs_problem(*train_dataset)
    params, opt_state = update(params, opt_state, problem)
    if step % 1000 == 0:
      train_loss = jnp.mean(
          jnp.asarray([
              accuracy_loss(params, get_higgs_problem(*train_dataset))
              for _ in range(100)
          ])).item()
      test_loss = jnp.mean(
          jnp.asarray([
              accuracy_loss(params, get_higgs_problem(*test_dataset))
              for _ in range(100)
          ])).item()
      logging.info("step %r loss train %r test %r", step, train_loss, test_loss)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")

  train(num_steps=10000)


if __name__ == "__main__":
  app.run(main)



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration file for the Sphinx documentation builder."""

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top
import inspect
import os
import sys
import typing
import jraph


def _add_annotations_import(path):
  """Appends a future annotations import to the file at the given path."""
  with open(path) as f:
    contents = f.read()
  # If we run sphinx multiple times then we will append the future import
  # multiple times too, so this check is here to prevent that.
  if contents.startswith('from __future__ import annotations'):
    return

  assert contents.startswith('#'), (path, contents.split('\n')[0])
  with open(path, 'w') as f:
    # NOTE: This is subtle and not unit tested, we're prefixing the first line
    # in each Python file with this future import. It is important to prefix
    # not insert a newline such that source code locations are accurate (we link
    # to GitHub). The assertion above ensures that the first line in the file is
    # a comment so it is safe to prefix it.
    f.write('from __future__ import annotations  ')
    f.write(contents)


def _recursive_add_annotations_import():
  for path, _, files in os.walk('../jraph/'):
    for file in files:
      if file.endswith('.py'):
        _add_annotations_import(os.path.abspath(os.path.join(path, file)))

if 'READTHEDOCS' in os.environ:
  _recursive_add_annotations_import()

typing.get_type_hints = lambda obj, *unused: obj.__annotations__
sys.path.insert(0, os.path.abspath('../'))
sys.path.append(os.path.abspath('ext'))

# -- Project information -----------------------------------------------------

project = 'Jraph'
copyright = '2021, Jraph Authors'  # pylint: disable=redefined-builtin
author = 'Jraph Authors'

# The full version, including alpha/beta/rc tags
release = '0.0.1.dev'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.linkcode',
    'sphinx.ext.napoleon',
    'sphinxcontrib.bibtex',
    'sphinx_autodoc_typehints',
]

pygments_style = 'sphinx'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build']

# -- Options for autodoc -----------------------------------------------------

autodoc_default_options = {
    'member-order': 'bysource',
    'special-members': True,
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Source code links -------------------------------------------------------


def linkcode_resolve(domain, info):
  """Resolve a GitHub URL corresponding to Python object."""
  if domain != 'py':
    return None

  try:
    mod = sys.modules[info['module']]
  except ImportError:
    return None

  obj = mod
  try:
    for attr in info['fullname'].split('.'):
      obj = getattr(obj, attr)
  except AttributeError:
    return None
  else:
    obj = inspect.unwrap(obj)

  try:
    filename = inspect.getsourcefile(obj)
  except TypeError:
    return None

  try:
    source, lineno = inspect.getsourcelines(obj)
  except OSError:
    return None

  return 'https://github.com/deepmind/jraph/blob/master/jraph/%s#L%d#L%d' % (
      os.path.relpath(filename, start=os.path.dirname(
          jraph.__file__)), lineno, lineno + len(source) - 1)



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph Data Structures."""

from typing import Any, NamedTuple, Iterable, Mapping, Union, Optional
import jax.numpy as jnp


# As of 04/2020 pytype doesn't support recursive types.
# pytype: disable=not-supported-yet
ArrayTree = Union[jnp.ndarray, Iterable['ArrayTree'], Mapping[Any, 'ArrayTree']]


class GraphsTuple(NamedTuple):
  """An ordered collection of graphs in a sparse format.

  The values of ``nodes``, ``edges`` and ``globals`` can be ``ArrayTrees`` -
  nests of features with ``jax`` compatible values. For example, ``nodes`` in a
  graph may have more than one type of attribute.

  However, the GraphsTuple typically takes the following form for a batch of
  `n` graphs:

  - n_node: The number of nodes per graph. It is a vector of integers with shape
    `[n_graphs]`, such that ``graph.n_node[i]`` is the number of nodes in the
    i-th graph.

  - n_edge: The number of edges per graph. It is a vector of integers with shape
    `[n_graphs]`, such that ``graph.n_edge[i]`` is the number of edges in the
    i-th graph.

  - nodes: The nodes features. It is either ``None`` (the graph has no node
    features), or a vector of shape `[n_nodes] + node_shape`, where
    ``n_nodes = sum(graph.n_node)`` is the total number of nodes in the batch of
    graphs, and `node_shape` represents the shape of the features of each node.
    The relative index of a node from the batched version can be recovered from
    the ``graph.n_node`` property. For instance, the second node of the third
    graph will have its features in the
    `1 + graph.n_node[0] + graph.n_node[1]`-th slot of graph.nodes.
    Observe that having a ``None`` value for this field does not mean that the
    graphs have no nodes, only that they do not have node features.

  - edges: The edges features. It is either ``None`` (the graph has no edge
    features), or a vector of shape `[n_edges] + edge_shape`, where
    ``n_edges = sum(graph.n_edge)`` is the total number of edges in the batch of
    graphs, and ``edge_shape`` represents the shape of the features of each
    edge.

    The relative index of an edge from the batched version can be recovered from
    the ``graph.n_edge`` property. For instance, the third edge of the third
    graph will have its features in the `2 + graph.n_edge[0] + graph.n_edge[1]`-
    th slot of graph.edges.

    Having a ``None`` value for this field does not necessarily mean that the
    graph has no edges, only that they do not have edge features.

  - receivers: The indices of the receiver nodes, for each edge. It is either
    ``None`` (if the graph has no edges), or a vector of integers of shape
    `[n_edges]`, such that ``graph.receivers[i]`` is the index of the node
    receiving from the i-th edge.

    Observe that the index is absolute (in other words, cumulative), i.e.
    ``graphs.receivers`` take value in `[0, n_nodes]`. For instance, an edge
    connecting the vertices with relative indices 2 and 3 in the second graph of
    the batch would have a ``receivers`` value of `3 + graph.n_node[0]`.
    If `graphs.receivers` is ``None``, then ``graphs.edges`` and
    ``graphs.senders`` should also be ``None``.

  - senders: The indices of the sender nodes, for each edge. It is either
    ``None`` (if the graph has no edges), or a vector of integers of shape
    `[n_edges]`, such that ``graph.senders[i]`` is the index of the node
    sending from the i-th edge.

    Observe that the index is absolute, i.e. ``graphs.senders`` take value in
    `[0, n_nodes]`. For instance, an edge connecting the vertices with relative
    indices 1 and 3 in the third graph of the batch would have a ``senders``
    value of `1 + graph.n_node[0] + graph.n_node[1]`.

    If ``graphs.senders`` is ``None``, then ``graphs.edges`` and
    ``graphs.receivers`` should also be ``None``.

  - globals: The global features of the graph. It is either ``None`` (the graph
    has no global features), or a vector of shape `[n_graphs] + global_shape`
    representing graph level features.


  """
  nodes: Optional[ArrayTree]
  edges: Optional[ArrayTree]
  receivers: Optional[jnp.ndarray]  # with integer dtype
  senders: Optional[jnp.ndarray]  # with integer dtype
  globals: Optional[ArrayTree]
  n_node: jnp.ndarray  # with integer dtype
  n_edge: jnp.ndarray   # with integer dtype



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data loading utils for the Open Graph Benchmark (OGB) Mol-Hiv."""

import pathlib
import jraph
import numpy as np
import pandas as pd


class DataReader:
  """Data Reader for Open Graph Benchmark datasets."""

  def __init__(self,
               data_path,
               master_csv_path,
               split_path,
               batch_size=1,
               dynamically_batch=False):
    """Initializes the data reader by loading in data."""
    with pathlib.Path(master_csv_path).open("rt") as fp:
      self._dataset_info = pd.read_csv(fp, index_col=0)["ogbg-molhiv"]
    self._data_path = pathlib.Path(data_path)
    # Load edge information, and transpose into (senders, receivers).
    with pathlib.Path(data_path, "edge.csv.gz").open("rb") as fp:
      sender_receivers = pd.read_csv(
          fp, compression="gzip", header=None).values.T.astype(np.int64)
      self._senders = sender_receivers[0]
      self._receivers = sender_receivers[1]
    # Load n_node and n_edge
    with pathlib.Path(data_path, "num-node-list.csv.gz").open("rb") as fp:
      self._n_node = pd.read_csv(fp, compression="gzip", header=None)
      self._n_node = self._n_node.astype(np.int64)[0].tolist()
    with pathlib.Path(data_path, "num-edge-list.csv.gz").open("rb") as fp:
      self._n_edge = pd.read_csv(fp, compression="gzip", header=None)
      self._n_edge = self._n_edge.astype(np.int64)[0].tolist()
    # Load node features
    with pathlib.Path(data_path, "node-feat.csv.gz").open("rb") as fp:
      self._nodes = pd.read_csv(
          fp, compression="gzip", header=None).astype(np.float32).values
    with pathlib.Path(data_path, "edge-feat.csv.gz").open("rb") as fp:
      self._edges = pd.read_csv(
          fp, compression="gzip", header=None).astype(np.float32).values
    with pathlib.Path(data_path, "graph-label.csv.gz").open("rb") as fp:
      self._labels = pd.read_csv(
          fp, compression="gzip", header=None).values

    with pathlib.Path(split_path).open("rb") as fp:
      self._split_idx = pd.read_csv(
          fp, compression="gzip", header=None).values.T[0]

    self._repeat = False
    self._batch_size = batch_size
    self._generator = self._make_generator()
    self._max_nodes = int(np.max(self._n_node))
    self._max_edges = int(np.max(self._n_edge))

    if dynamically_batch:
      self._generator = jraph.dynamically_batch(
          self._generator,
          # Plus one for the extra padding node.
          n_node=self._batch_size * (self._max_nodes) + 1,
          # Times two because we want backwards edges.
          n_edge=self._batch_size * (self._max_edges) * 2,
          n_graph=self._batch_size + 1)

    # If n_node = [1,2,3], we create accumulated n_node [0,1,3,6] for indexing.
    self._accumulated_n_nodes = np.concatenate((np.array([0]),
                                                np.cumsum(self._n_node)))
    # Same for n_edge
    self._accumulated_n_edges = np.concatenate((np.array([0]),
                                                np.cumsum(self._n_edge)))
    self._dynamically_batch = dynamically_batch

  @property
  def total_num_graphs(self):
    return len(self._n_node)

  def repeat(self):
    self._repeat = True

  def __iter__(self):
    return self

  def __next__(self):
    graphs = []
    if self._dynamically_batch:
      # If we are using pmap we need each batch to have the same size in both
      # number of nodes and number of edges. So we use dynamically batch which
      # guarantees this.
      return next(self._generator)
    else:
      for _ in range(self._batch_size):
        graph = next(self._generator)
        graphs.append(graph)
      return jraph.batch(graphs)

  def get_graph_by_idx(self, idx):
    """Gets a graph by an integer index."""
    # Gather the graph information
    label = self._labels[idx]
    n_node = self._n_node[idx]
    n_edge = self._n_edge[idx]
    node_slice = slice(
        self._accumulated_n_nodes[idx], self._accumulated_n_nodes[idx+1])
    edge_slice = slice(
        self._accumulated_n_edges[idx], self._accumulated_n_edges[idx+1])
    nodes = self._nodes[node_slice]
    edges = self._edges[edge_slice]
    senders = self._senders[edge_slice]
    receivers = self._receivers[edge_slice]
    # Molecular graphs are bi directional, but the serialization only
    # stores one way so we add the missing edges.
    return jraph.GraphsTuple(
        nodes=nodes,
        edges=np.concatenate([edges, edges]),
        n_node=np.array([n_node]),
        n_edge=np.array([n_edge*2]),
        senders=np.concatenate([senders, receivers]),
        receivers=np.concatenate([receivers, senders]),
        globals={"label": label})

  def _make_generator(self):
    """Makes a single example generator of the loaded OGB data."""
    idx = 0
    while True:
      # If not repeating, exit when we've cycled through all the graphs.
      # Only return graphs within the split.
      if not self._repeat:
        if idx == self.total_num_graphs:
          return
      else:
        # This will reset the index to 0 if we are at the end of the dataset.
        idx = idx % self.total_num_graphs
      if idx not in self._split_idx:
        idx += 1
        continue
      graph = self.get_graph_by_idx(idx)
      idx += 1
      yield graph



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Jraph."""


from jraph._src.graph import GraphsTuple
from jraph._src.models import AggregateEdgesToGlobalsFn
from jraph._src.models import AggregateEdgesToNodesFn
from jraph._src.models import AggregateNodesToGlobalsFn
from jraph._src.models import AttentionLogitFn
from jraph._src.models import AttentionReduceFn
from jraph._src.models import DeepSets
from jraph._src.models import EmbedEdgeFn
from jraph._src.models import EmbedGlobalFn
from jraph._src.models import EmbedNodeFn
from jraph._src.models import GAT
from jraph._src.models import GATAttentionLogitFn
from jraph._src.models import GATAttentionQueryFn
from jraph._src.models import GATNodeUpdateFn
from jraph._src.models import GNUpdateEdgeFn
from jraph._src.models import GNUpdateGlobalFn
from jraph._src.models import GNUpdateNodeFn
from jraph._src.models import GraphConvolution
from jraph._src.models import GraphMapFeatures
from jraph._src.models import GraphNetGAT
from jraph._src.models import GraphNetwork
from jraph._src.models import InteractionNetwork
from jraph._src.models import InteractionUpdateEdgeFn
from jraph._src.models import InteractionUpdateNodeFn
from jraph._src.models import NodeFeatures
from jraph._src.models import RelationNetwork
from jraph._src.utils import ArrayTree
from jraph._src.utils import batch
from jraph._src.utils import batch_np
from jraph._src.utils import concatenated_args
from jraph._src.utils import dynamically_batch
from jraph._src.utils import get_edge_padding_mask
from jraph._src.utils import get_fully_connected_graph
from jraph._src.utils import get_graph_padding_mask
from jraph._src.utils import get_node_padding_mask
from jraph._src.utils import get_number_of_padding_with_graphs_edges
from jraph._src.utils import get_number_of_padding_with_graphs_graphs
from jraph._src.utils import get_number_of_padding_with_graphs_nodes
from jraph._src.utils import pad_with_graphs
from jraph._src.utils import partition_softmax
from jraph._src.utils import segment_max
from jraph._src.utils import segment_max_or_constant
from jraph._src.utils import segment_mean
from jraph._src.utils import segment_min
from jraph._src.utils import segment_min_or_constant
from jraph._src.utils import segment_normalize
from jraph._src.utils import segment_softmax
from jraph._src.utils import segment_sum
from jraph._src.utils import segment_variance
from jraph._src.utils import unbatch
from jraph._src.utils import unbatch_np
from jraph._src.utils import unpad_with_graphs
from jraph._src.utils import with_zero_out_padding_outputs
from jraph._src.utils import zero_out_padding


__version__ = "0.0.6.dev0"

__all__ = ("ArrayTree", "DeepSets", "GraphConvolution", "GraphMapFeatures",
           "InteractionNetwork", "RelationNetwork", "GraphNetGAT", "GAT",
           "GraphsTuple", "GraphNetwork", "NodeFeatures",
           "AggregateEdgesToNodesFn", "AggregateNodesToGlobalsFn",
           "AggregateEdgesToGlobalsFn", "AttentionLogitFn", "AttentionReduceFn",
           "GNUpdateEdgeFn", "GNUpdateNodeFn", "GNUpdateGlobalFn",
           "InteractionUpdateNodeFn", "InteractionUpdateEdgeFn", "EmbedEdgeFn",
           "EmbedNodeFn", "EmbedGlobalFn", "GATAttentionQueryFn",
           "GATAttentionLogitFn", "GATNodeUpdateFn", "batch", "batch_np",
           "unbatch", "unbatch_np", "pad_with_graphs",
           "get_number_of_padding_with_graphs_graphs",
           "get_number_of_padding_with_graphs_nodes",
           "get_number_of_padding_with_graphs_edges", "unpad_with_graphs",
           "get_node_padding_mask", "get_edge_padding_mask",
           "get_graph_padding_mask", "segment_max", "segment_max_or_constant",
           "segment_min_or_constant", "segment_softmax", "segment_sum",
           "partition_softmax", "concatenated_args",
           "get_fully_connected_graph", "dynamically_batch",
           "with_zero_out_padding_outputs", "zero_out_padding")

#  _________________________________________
# / Please don't use symbols in `_src` they \
# \ are not part of the Jraph public API.    /
#  -----------------------------------------
#         \   ^__^
#          \  (oo)\_______
#             (__)\       )\/\
#                 ||----w |
#                 ||     ||
#
try:
  del _src  # pylint: disable=undefined-variable
except NameError:
  pass



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Example of how to use recurrent networks (e.g.`LSTM`s) with `GraphNetwork`s.

Models can use the mechanism for specifying nested node, edge, or global
features to simultaneously keep inputs/embeddings together with a per-node,
per-edge or per-graph recurrent state.

In this example we show an `InteractionNetwork` that uses an LSTM to keep a
memory of the inputs to the edge model at each step of message passing, by using
separate "embedding" and "state" fields in the edge features.
Following a similar procedure, an LSTM could be added to the `node_update_fn`,
or even the `global_update_fn`, if using a full `GraphNetwork`.

Note it is recommended to use immutable container types to store nested edge,
node and global features to avoid unwanted side effects. In this example we
use `namedtuple`s.

"""

import collections

from absl import app
import haiku as hk
import jax
import jax.numpy as jnp
import jax.tree_util as tree
import jraph
import numpy as np


NUM_NODES = 5
NUM_EDGES = 7
NUM_MESSAGE_PASSING_STEPS = 10
EMBEDDING_SIZE = 32
HIDDEN_SIZE = 128

# Immutable class for storing nested node/edge features containing an embedding
# and a recurrent state.
StatefulField = collections.namedtuple("StatefulField", ["embedding", "state"])


def get_random_graph() -> jraph.GraphsTuple:
  return jraph.GraphsTuple(
      n_node=np.asarray([NUM_NODES]),
      n_edge=np.asarray([NUM_EDGES]),
      nodes=np.random.normal(size=[NUM_NODES, EMBEDDING_SIZE]),
      edges=np.random.normal(size=[NUM_EDGES, EMBEDDING_SIZE]),
      globals=None,
      senders=np.random.randint(0, NUM_NODES, [NUM_EDGES]),
      receivers=np.random.randint(0, NUM_NODES, [NUM_EDGES]))


def network_definition(graph: jraph.GraphsTuple) -> jraph.ArrayTree:
  """`InteractionNetwork` with an LSTM in the edge update."""

  # LSTM that will keep a memory of the inputs to the edge model.
  edge_fn_lstm = hk.LSTM(hidden_size=HIDDEN_SIZE)

  # MLPs used in the edge and the node model. Note that in this instance
  # the output size matches the input size so the same model can be run
  # iteratively multiple times. In a real model, this would usually be achieved
  # by first using an encoder in the input data into a common `EMBEDDING_SIZE`.
  edge_fn_mlp = hk.nets.MLP([HIDDEN_SIZE, EMBEDDING_SIZE])
  node_fn_mlp = hk.nets.MLP([HIDDEN_SIZE, EMBEDDING_SIZE])

  # Initialize the edge features to contain both the input edge embedding
  # and initial LSTM state. Note for the nodes we only have an embedding since
  # in this example nodes do not use a `node_fn_lstm`, but for analogy, we
  # still put it in a `StatefulField`.
  graph = graph._replace(
      edges=StatefulField(
          embedding=graph.edges,
          state=edge_fn_lstm.initial_state(graph.edges.shape[0])),
      nodes=StatefulField(embedding=graph.nodes, state=None),
  )

  def update_edge_fn(edges, sender_nodes, receiver_nodes):
    # We will run an LSTM memory on the inputs first, and then
    # process the output of the LSTM with an MLP.
    edge_inputs = jnp.concatenate([edges.embedding,
                                   sender_nodes.embedding,
                                   receiver_nodes.embedding], axis=-1)
    lstm_output, updated_state = edge_fn_lstm(edge_inputs, edges.state)
    updated_edges = StatefulField(
        embedding=edge_fn_mlp(lstm_output), state=updated_state,
    )
    return updated_edges

  def update_node_fn(nodes, received_edges):
    # Note `received_edges.state` will also contain the aggregated state for
    # all received edges, which we may choose to use in the node update.
    node_inputs = jnp.concatenate(
        [nodes.embedding, received_edges.embedding], axis=-1)
    updated_nodes = StatefulField(
        embedding=node_fn_mlp(node_inputs),
        state=None)
    return updated_nodes

  recurrent_graph_network = jraph.InteractionNetwork(
      update_edge_fn=update_edge_fn,
      update_node_fn=update_node_fn)

  # Apply the model recurrently for 10 message passing steps.
  # If instead we intended to use the LSTM to process a sequence of features
  # for each node/edge, here we would select the corresponding inputs from the
  # sequence along the sequence axis of the nodes/edges features to build the
  # correct input graph for each step of the iteration.
  num_message_passing_steps = 10
  for _ in range(num_message_passing_steps):
    graph = recurrent_graph_network(graph)

  return graph


def main(_):

  network = hk.without_apply_rng(hk.transform(network_definition))
  input_graph = get_random_graph()
  params = network.init(jax.random.PRNGKey(42), input_graph)
  output_graph = network.apply(params, input_graph)
  print(tree.tree_map(lambda x: x.shape, output_graph))


if __name__ == "__main__":
  app.run(main)



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for graph.ogb_examples.data_utils."""

import pathlib
from absl.testing import absltest
from absl.testing import parameterized
import jraph
from jraph.ogb_examples import data_utils
import numpy as np
import tree


class DataUtilsTest(parameterized.TestCase):

  def setUp(self):
    super(DataUtilsTest, self).setUp()
    self._test_graph = jraph.GraphsTuple(
        nodes=np.broadcast_to(
            np.arange(10, dtype=np.float32)[:, None], (10, 10)),
        edges=np.concatenate((
            np.broadcast_to(np.arange(20, dtype=np.float32)[:, None], (20, 4)),
            np.broadcast_to(np.arange(20, dtype=np.float32)[:, None], (20, 4))
            )),
        receivers=np.concatenate((np.arange(20), np.arange(20))),
        senders=np.concatenate((np.arange(20), np.arange(20))),
        globals={'label': np.array([1], dtype=np.int32)},
        n_node=np.array([10], dtype=np.int32),
        n_edge=np.array([40], dtype=np.int32))
    ogb_path = pathlib.Path(data_utils.__file__).parents[0]
    master_csv_path = pathlib.Path(ogb_path, 'test_data', 'master.csv')
    split_path = pathlib.Path(ogb_path, 'test_data', 'train.csv.gz')
    data_path = master_csv_path.parents[0]
    self._reader = data_utils.DataReader(
        data_path=data_path,
        master_csv_path=master_csv_path,
        split_path=split_path)

  def test_total_num_graph(self):
    self.assertEqual(self._reader.total_num_graphs, 1)

  def test_expected_graph(self):
    graph = next(self._reader)
    with self.subTest('test_graph_equality'):
      tree.map_structure(
          np.testing.assert_almost_equal, graph, self._test_graph)
    with self.subTest('stop_iteration'):
      # One element in the dataset, so should have stop iteration.
      with self.assertRaises(StopIteration):
        next(self._reader)

  def test_reader_repeat(self):
    self._reader.repeat()
    next(self._reader)
    graph = next(self._reader)
    # One graph in the test dataset so should be the same.
    tree.map_structure(np.testing.assert_almost_equal, graph, self._test_graph)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""A basic graphnet example.

This example just explains the bare mechanics of the library.
"""

import logging

from absl import app
import jax
import jraph
import numpy as np

MASK_BROKEN_MSG = ("Support for jax.mask is currently broken. This is not a "
                   "jraph error.")


def run():
  """Runs basic example."""

  # Creating graph tuples.

  # Creates a GraphsTuple from scratch containing a single graph.
  # The graph has 3 nodes and 2 edges.
  # Each node has a 4-dimensional feature vector.
  # Each edge has a 5-dimensional feature vector.
  # The graph itself has a 6-dimensional feature vector.
  single_graph = jraph.GraphsTuple(
      n_node=np.asarray([3]), n_edge=np.asarray([2]),
      nodes=np.ones((3, 4)), edges=np.ones((2, 5)),
      globals=np.ones((1, 6)),
      senders=np.array([0, 1]), receivers=np.array([2, 2]))
  logging.info("Single graph %r", single_graph)

  # Creates a GraphsTuple from scratch containing a single graph with nested
  # feature vectors.
  # The graph has 3 nodes and 2 edges.
  # The feature vector can be arbitrary nested types of dict, list and tuple,
  # or any other type you registered with jax.tree_util.register_pytree_node.
  nested_graph = jraph.GraphsTuple(
      n_node=np.asarray([3]), n_edge=np.asarray([2]),
      nodes={"a": np.ones((3, 4))}, edges={"b": np.ones((2, 5))},
      globals={"c": np.ones((1, 6))},
      senders=np.array([0, 1]), receivers=np.array([2, 2]))
  logging.info("Nested graph %r", nested_graph)

  # Creates a GraphsTuple from scratch containing 2 graphs using an implicit
  # batch dimension.
  # The first graph has 3 nodes and 2 edges.
  # The second graph has 1 node and 1 edge.
  # Each node has a 4-dimensional feature vector.
  # Each edge has a 5-dimensional feature vector.
  # The graph itself has a 6-dimensional feature vector.
  implicitly_batched_graph = jraph.GraphsTuple(
      n_node=np.asarray([3, 1]), n_edge=np.asarray([2, 1]),
      nodes=np.ones((4, 4)), edges=np.ones((3, 5)),
      globals=np.ones((2, 6)),
      senders=np.array([0, 1, 3]), receivers=np.array([2, 2, 3]))
  logging.info("Implicitly batched graph %r", implicitly_batched_graph)

  # Batching graphs can be challenging. There are in general two approaches:
  # 1. Implicit batching: Independent graphs are combined into the same
  #    GraphsTuple first, and the padding is added to the combined graph.
  # 2. Explicit batching: Pad all graphs to a maximum size, stack them together
  #    using an explicit batch dimension followed by jax.vmap.
  # Both approaches are shown below.

  # Creates a GraphsTuple from two existing GraphsTuple using an implicit
  # batch dimension.
  # The GraphsTuple will contain three graphs.
  implicitly_batched_graph = jraph.batch(
      [single_graph, implicitly_batched_graph])
  logging.info("Implicitly batched graph %r", implicitly_batched_graph)

  # Creates multiple GraphsTuples from an existing GraphsTuple with an implicit
  # batch dimension.
  graph_1, graph_2, graph_3 = jraph.unbatch(implicitly_batched_graph)
  logging.info("Unbatched graphs %r %r %r", graph_1, graph_2, graph_3)

  # Creates a padded GraphsTuple from an existing GraphsTuple.
  # The padded GraphsTuple will contain 10 nodes, 5 edges, and 4 graphs.
  # Three graphs are added for the padding.
  # First a dummy graph which contains the padding nodes and edges and secondly
  # two empty graphs without nodes or edges to pad out the graphs.
  padded_graph = jraph.pad_with_graphs(
      single_graph, n_node=10, n_edge=5, n_graph=4)
  logging.info("Padded graph %r", padded_graph)

  # Creates a GraphsTuple from an existing padded GraphsTuple.
  # The previously added padding is removed.
  single_graph = jraph.unpad_with_graphs(padded_graph)
  logging.info("Unpadded graph %r", single_graph)

  # Creates a GraphsTuple containing 2 graphs using an explicit batch
  # dimension.
  # An explicit batch dimension requires more memory, but can simplify
  # the definition of functions operating on the graph.
  # Explicitly batched graphs require the GraphNetwork to be transformed
  # by jax.vmap.
  # Using an explicit batch requires padding all feature vectors to
  # the maximum size of nodes and edges.
  # The first graph has 3 nodes and 2 edges.
  # The second graph has 1 node and 1 edge.
  # Each node has a 4-dimensional feature vector.
  # Each edge has a 5-dimensional feature vector.
  # The graph itself has a 6-dimensional feature vector.
  explicitly_batched_graph = jraph.GraphsTuple(
      n_node=np.asarray([[3], [1]]), n_edge=np.asarray([[2], [1]]),
      nodes=np.ones((2, 3, 4)), edges=np.ones((2, 2, 5)),
      globals=np.ones((2, 1, 6)),
      senders=np.array([[0, 1], [0, -1]]),
      receivers=np.array([[2, 2], [0, -1]]))
  logging.info("Explicitly batched graph %r", explicitly_batched_graph)

  # Running a graph propagation step.
  # First define the update functions for the edges, nodes and globals.
  # In this example we use the identity everywhere.
  # For Graph neural networks, each update function is typically a neural
  # network.
  def update_edge_fn(
      edge_features,
      sender_node_features,
      receiver_node_features,
      globals_):
    """Returns the update edge features."""
    del sender_node_features
    del receiver_node_features
    del globals_
    return edge_features

  def update_node_fn(
      node_features,
      aggregated_sender_edge_features,
      aggregated_receiver_edge_features,
      globals_):
    """Returns the update node features."""
    del aggregated_sender_edge_features
    del aggregated_receiver_edge_features
    del globals_
    return node_features

  def update_globals_fn(
      aggregated_node_features,
      aggregated_edge_features,
      globals_):
    """Returns the global features."""
    del aggregated_node_features
    del aggregated_edge_features
    return globals_

  # Optionally define custom aggregation functions.
  # In this example we use the defaults (so no need to define them explicitly).
  aggregate_edges_for_nodes_fn = jraph.segment_sum
  aggregate_nodes_for_globals_fn = jraph.segment_sum
  aggregate_edges_for_globals_fn = jraph.segment_sum

  # Optionally define an attention logit function and an attention reduce
  # function. This can be used for graph attention.
  # The attention function calculates attention weights, and the apply
  # attention function calculates the new edge feature given the weights.
  # We don't use graph attention here, and just pass the defaults.
  attention_logit_fn = None
  attention_reduce_fn = None

  # Creates a new GraphNetwork in its most general form.
  # Most of the arguments have defaults and can be omitted if a feature
  # is not used.
  # There are also predefined GraphNetworks available (see models.py)
  network = jraph.GraphNetwork(
      update_edge_fn=update_edge_fn,
      update_node_fn=update_node_fn,
      update_global_fn=update_globals_fn,
      attention_logit_fn=attention_logit_fn,
      aggregate_edges_for_nodes_fn=aggregate_edges_for_nodes_fn,
      aggregate_nodes_for_globals_fn=aggregate_nodes_for_globals_fn,
      aggregate_edges_for_globals_fn=aggregate_edges_for_globals_fn,
      attention_reduce_fn=attention_reduce_fn)

  # Runs graph propagation on (implicitly batched) graphs.
  updated_graph = network(single_graph)
  logging.info("Updated graph from single graph %r", updated_graph)

  updated_graph = network(nested_graph)
  logging.info("Updated graph from nested graph %r", nested_graph)

  updated_graph = network(implicitly_batched_graph)
  logging.info("Updated graph from implicitly batched graph %r", updated_graph)

  updated_graph = network(padded_graph)
  logging.info("Updated graph from padded graph %r", updated_graph)

  # JIT-compile graph propagation.
  # Use padded graphs to avoid re-compilation at every step!
  jitted_network = jax.jit(network)
  updated_graph = jitted_network(padded_graph)
  logging.info("(JIT) updated graph from padded graph %r", updated_graph)
  logging.info("basic.py complete!")


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")
  run()


if __name__ == "__main__":
  app.run(main)



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Example training script for training OGB molhiv with jax graph-nets.

The ogbg-molhiv dataset is a molecular property prediction dataset.
It is adopted from the MoleculeNet [1]. All the molecules are pre-processed
using RDKit [2].

Each graph represents a molecule, where nodes are atoms, and edges are chemical
bonds. Input node features are 9-dimensional, containing atomic number and
chirality, as well as other additional atom features such as formal charge and
whether the atom is in the ring or not.

The goal is to predict whether a molecule inhibits HIV virus replication or not.
Performance is measured in ROC-AUC.

This script uses a GraphNet to learn the prediction task.

[1] Zhenqin Wu, Bharath Ramsundar, Evan N Feinberg, Joseph Gomes,
Caleb Geniesse, Aneesh SPappu, Karl Leswing, and Vijay Pande.
Moleculenet: a benchmark for molecular machine learning.
Chemical Science, 9(2):513–530, 2018.

[2] Greg Landrum et al. RDKit: Open-source cheminformatics, 2006.

Example usage:

python3 train.py --data_path={DATA_PATH} --master_csv_path={MASTER_CSV_PATH} \
--save_dir={SAVE_DIR} --split_path={SPLIT_PATH}
"""

import functools
import logging
import pathlib
import pickle
from typing import Iterator
from absl import app
from absl import flags
import haiku as hk
import jax
import jax.numpy as jnp
import jraph
from jraph.ogb_examples import data_utils
import optax


flags.DEFINE_string('data_path', None, 'Directory of the data.')
flags.DEFINE_string('split_path', None, 'Path to the data split indices.')
flags.DEFINE_string('master_csv_path', None, 'Path to OGB master.csv.')
flags.DEFINE_string('save_dir', None, 'Directory to save parameters to.')
flags.DEFINE_integer('batch_size', 1, 'Number of graphs in batch.')
flags.DEFINE_integer('num_training_steps', 1000, 'Number of training steps.')
flags.DEFINE_enum('mode', 'train', ['train', 'evaluate'], 'Train or evaluate.')
FLAGS = flags.FLAGS


@jraph.concatenated_args
def edge_update_fn(feats: jnp.ndarray) -> jnp.ndarray:
  """Edge update function for graph net."""
  net = hk.Sequential(
      [hk.Linear(128), jax.nn.relu,
       hk.Linear(128)])
  return net(feats)


@jraph.concatenated_args
def node_update_fn(feats: jnp.ndarray) -> jnp.ndarray:
  """Node update function for graph net."""
  net = hk.Sequential(
      [hk.Linear(128), jax.nn.relu,
       hk.Linear(128)])
  return net(feats)


@jraph.concatenated_args
def update_global_fn(feats: jnp.ndarray) -> jnp.ndarray:
  """Global update function for graph net."""
  # Molhiv is a binary classification task, so output pos neg logits.
  net = hk.Sequential(
      [hk.Linear(128), jax.nn.relu,
       hk.Linear(2)])
  return net(feats)


def net_fn(graph: jraph.GraphsTuple) -> jraph.GraphsTuple:
  """Graph net function."""
  # Add a global paramater for graph classification.
  graph = graph._replace(globals=jnp.zeros([graph.n_node.shape[0], 1]))
  embedder = jraph.GraphMapFeatures(
      hk.Linear(128), hk.Linear(128), hk.Linear(128))
  net = jraph.GraphNetwork(
      update_node_fn=node_update_fn,
      update_edge_fn=edge_update_fn,
      update_global_fn=update_global_fn)
  return net(embedder(graph))


def device_batch(
    graph_generator: data_utils.DataReader) -> Iterator[jraph.GraphsTuple]:
  """Batches a set of graphs the size of the number of devices."""
  num_devices = jax.local_device_count()
  batch = []
  for idx, graph in enumerate(graph_generator):
    if idx % num_devices == num_devices - 1:
      batch.append(graph)
      yield jax.tree_map(lambda *x: jnp.stack(x, axis=0), *batch)
      batch = []
    else:
      batch.append(graph)


def compute_loss(params, graph, label, net):
  """Computes loss."""
  pred_graph = net.apply(params, graph)
  preds = jax.nn.log_softmax(pred_graph.globals)
  targets = jax.nn.one_hot(label, 2)

  # Since we have an extra 'dummy' graph in our batch due to padding, we want
  # to mask out any loss associated with the dummy graph.
  # Since we padded with `pad_with_graphs` we can recover the mask by using
  # get_graph_padding_mask.
  mask = jraph.get_graph_padding_mask(pred_graph)

  # Cross entropy loss.
  loss = -jnp.mean(preds * targets * mask[:, None])

  # Accuracy taking into account the mask.
  accuracy = jnp.sum(
      (jnp.argmax(pred_graph.globals, axis=1) == label) * mask)/jnp.sum(mask)
  return loss, accuracy


def train(data_path, master_csv_path, split_path, batch_size,
          num_training_steps, save_dir):
  """OGB Training Script."""
  # Initialize the dataset reader.
  reader = data_utils.DataReader(
      data_path=data_path,
      master_csv_path=master_csv_path,
      split_path=split_path,
      batch_size=batch_size,
      dynamically_batch=True)
  # Repeat the dataset forever for training.
  reader.repeat()

  # Transform impure `net_fn` to pure functions with hk.transform.
  net = hk.without_apply_rng(hk.transform(net_fn))
  # Get a candidate graph and label to initialize the network.
  graph = reader.get_graph_by_idx(0)

  # Initialize the network.
  logging.info('Initializing network.')
  params = net.init(jax.random.PRNGKey(42), graph)
  # Because we are training with multiple devices, params needs to have a
  # device axis.
  params = jax.device_put_replicated(params, list(jax.devices()))
  # Initialize the optimizer.
  opt_init, opt_update = optax.adam(1e-4)
  opt_state = jax.pmap(opt_init)(params)

  compute_loss_fn = functools.partial(compute_loss, net=net)
  # We pmap the computation of our loss, since this is the main computation.
  # Using jax.pmap means that we will use all available accelerators.
  # More information can be found in the jax documentation.
  @functools.partial(jax.pmap, axis_name='device')
  def update_fn(params, graph, label, opt_state):
    (loss, acc), grad = jax.value_and_grad(
        compute_loss_fn, has_aux=True)(params, graph, label)
    # Average gradients across devices
    grad = jax.lax.pmean(grad, axis_name='device')
    updates, opt_state = opt_update(grad, opt_state, params)
    params = optax.apply_updates(params, updates)
    return loss, acc, opt_state, params

  for idx in range(num_training_steps):
    graph_batch = next(device_batch(reader))
    label = graph_batch.globals['label']
    loss, acc, opt_state, params = update_fn(
        params, graph_batch, label, opt_state)
    if idx % 100 == 0:
      logging.info('step: %s, loss: %s, acc: %s', idx, loss, acc)
  if save_dir is not None:
    with pathlib.Path(save_dir, 'molhiv.pkl').open('wb') as fp:
      logging.info('Saving model to %s', save_dir)
      pickle.dump(params, fp)
  logging.info('Training finished')


def evaluate(data_path, master_csv_path, split_path, save_dir):
  """Evaluation Script."""
  logging.info('Evaluating OGB molviv')
  logging.info('Dataset split: %s', split_path)
  # Initialize the dataset reader.
  reader = data_utils.DataReader(
      data_path=data_path,
      master_csv_path=master_csv_path,
      split_path=split_path,
      batch_size=1,
      dynamically_batch=True)
  # Transform impure `net_fn` to pure functions with hk.transform.
  net = hk.without_apply_rng(hk.transform(net_fn))
  with pathlib.Path(save_dir, 'molhiv.pkl').open('rb') as fp:
    params = pickle.load(fp)
  accumulated_loss = 0
  accumulated_accuracy = 0
  idx = 0

  # We pmap the computation of our loss, since this is the main computation.
  compute_loss_fn = jax.pmap(functools.partial(compute_loss, net=net))
  for graph_batch in device_batch(reader):
    label = graph_batch.globals['label']
    loss, acc = compute_loss_fn(params, graph_batch, label)
    accumulated_accuracy += jnp.sum(acc)
    accumulated_loss += jnp.sum(loss)
    total_num_padding_graphs = jnp.sum(
        jax.vmap(jraph.get_number_of_padding_with_graphs_graphs)(graph_batch))
    idx += graph_batch.n_node.size - total_num_padding_graphs
    if idx % 100 == 0:
      logging.info('Evaluated %s graphs', idx)
    logging.info('Completed evaluation.')
  loss = accumulated_loss / idx
  accuracy = accumulated_accuracy / idx
  logging.info('Eval loss: %s, accuracy %s', loss, accuracy)
  return loss, accuracy


def main(_):
  if FLAGS.mode == 'train':
    train(FLAGS.data_path, FLAGS.master_csv_path, FLAGS.split_path,
          FLAGS.batch_size, FLAGS.num_training_steps, FLAGS.save_dir)
  elif FLAGS.mode == 'evaluate':
    evaluate(FLAGS.data_path, FLAGS.master_csv_path, FLAGS.split_path,
             FLAGS.save_dir)

if __name__ == '__main__':
  app.run(main)



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Example training script for training OGB molhiv with jax graph-nets & flax.

The ogbg-molhiv dataset is a molecular property prediction dataset.
It is adopted from the MoleculeNet [1]. All the molecules are pre-processed
using RDKit [2].

Each graph represents a molecule, where nodes are atoms, and edges are chemical
bonds. Input node features are 9-dimensional, containing atomic number and
chirality, as well as other additional atom features such as formal charge and
whether the atom is in the ring or not.

The goal is to predict whether a molecule inhibits HIV virus replication or not.
Performance is measured in ROC-AUC.

This script uses a GraphNet to learn the prediction task.

[1] Zhenqin Wu, Bharath Ramsundar, Evan N Feinberg, Joseph Gomes,
Caleb Geniesse, Aneesh SPappu, Karl Leswing, and Vijay Pande.
Moleculenet: a benchmark for molecular machine learning.
Chemical Science, 9(2):513–530, 2018.

[2] Greg Landrum et al. RDKit: Open-source cheminformatics, 2006.

Example usage:

python3 train.py --data_path={DATA_PATH} --master_csv_path={MASTER_CSV_PATH} \
--save_dir={SAVE_DIR} --split_path={SPLIT_PATH}
"""


import functools
import logging
import pathlib
import pickle
from typing import Sequence
from absl import app
from absl import flags
from flax import linen as nn
from flax import optim
import jax
import jax.numpy as jnp
import jraph
from jraph.ogb_examples import data_utils


flags.DEFINE_string('data_path', None, 'Directory of the data.')
flags.DEFINE_string('split_path', None, 'Path to the data split indices.')
flags.DEFINE_string('master_csv_path', None, 'Path to OGB master.csv.')
flags.DEFINE_string('save_dir', None, 'Directory to save parameters to.')
flags.DEFINE_integer('batch_size', 1, 'Number of graphs in batch.')
flags.DEFINE_integer('num_training_steps', 1000, 'Number of training steps.')
flags.DEFINE_enum('mode', 'train', ['train', 'evaluate'], 'Train or evaluate.')
FLAGS = flags.FLAGS


class ExplicitMLP(nn.Module):
  """A flax MLP."""
  features: Sequence[int]

  @nn.compact
  def __call__(self, inputs):
    x = inputs
    for i, lyr in enumerate([nn.Dense(feat) for feat in self.features]):
      x = lyr(x)
      if i != len(self.features) - 1:
        x = nn.relu(x)
    return x


# Functions must be passed to jraph GNNs, but pytype does not recognise
# linen Modules as callables to here we wrap in a function.
def make_embed_fn(latent_size):
  def embed(inputs):
    return nn.Dense(latent_size)(inputs)
  return embed


def make_mlp(features):
  @jraph.concatenated_args
  def update_fn(inputs):
    return ExplicitMLP(features)(inputs)
  return update_fn


class GraphNetwork(nn.Module):
  """A flax GraphNetwork."""
  mlp_features: Sequence[int]
  latent_size: int

  @nn.compact
  def __call__(self, graph):
    # Add a global parameter for graph classification.
    graph = graph._replace(globals=jnp.zeros([graph.n_node.shape[0], 1]))

    embedder = jraph.GraphMapFeatures(
        embed_node_fn=make_embed_fn(self.latent_size),
        embed_edge_fn=make_embed_fn(self.latent_size),
        embed_global_fn=make_embed_fn(self.latent_size))
    net = jraph.GraphNetwork(
        update_node_fn=make_mlp(self.mlp_features),
        update_edge_fn=make_mlp(self.mlp_features),
        # The global update outputs size 2 for binary classification.
        update_global_fn=make_mlp(self.mlp_features + (2,)))  # pytype: disable=unsupported-operands
    return net(embedder(graph))


def _nearest_bigger_power_of_two(x: int) -> int:
  """Computes the nearest power of two greater than x for padding."""
  y = 2
  while y < x:
    y *= 2
  return y


def pad_graph_to_nearest_power_of_two(
    graphs_tuple: jraph.GraphsTuple) -> jraph.GraphsTuple:
  """Pads a batched `GraphsTuple` to the nearest power of two.

  For example, if a `GraphsTuple` has 7 nodes, 5 edges and 3 graphs, this method
  would pad the `GraphsTuple` nodes and edges:
    7 nodes --> 8 nodes (2^3)
    5 edges --> 8 edges (2^3)

  And since padding is accomplished using `jraph.pad_with_graphs`, an extra
  graph and node is added:
    8 nodes --> 9 nodes
    3 graphs --> 4 graphs

  Args:
    graphs_tuple: a batched `GraphsTuple` (can be batch size 1).

  Returns:
    A graphs_tuple batched to the nearest power of two.
  """
  # Add 1 since we need at least one padding node for pad_with_graphs.
  pad_nodes_to = _nearest_bigger_power_of_two(jnp.sum(graphs_tuple.n_node)) + 1
  pad_edges_to = _nearest_bigger_power_of_two(jnp.sum(graphs_tuple.n_edge))
  # Add 1 since we need at least one padding graph for pad_with_graphs.
  # We do not pad to nearest power of two because the batch size is fixed.
  pad_graphs_to = graphs_tuple.n_node.shape[0] + 1
  return jraph.pad_with_graphs(graphs_tuple, pad_nodes_to, pad_edges_to,
                               pad_graphs_to)


def compute_loss(params, graph, label, net):
  """Computes loss."""
  pred_graph = net.apply(params, graph)
  preds = jax.nn.log_softmax(pred_graph.globals)
  targets = jax.nn.one_hot(label, 2)

  # Since we have an extra 'dummy' graph in our batch due to padding, we want
  # to mask out any loss associated with the dummy graph.
  # Since we padded with `pad_with_graphs` we can recover the mask by using
  # get_graph_padding_mask.
  mask = jraph.get_graph_padding_mask(pred_graph)

  # Cross entropy loss.
  loss = -jnp.mean(preds * targets * mask[:, None])

  # Accuracy taking into account the mask.
  accuracy = jnp.sum(
      (jnp.argmax(pred_graph.globals, axis=1) == label) * mask)/jnp.sum(mask)
  return loss, accuracy


def train_step(optimizer, graph, label, net):
  partial_loss_fn = functools.partial(
      compute_loss, graph=graph, label=label, net=net)
  grad_fn = jax.value_and_grad(partial_loss_fn, has_aux=True)
  (loss, accuracy), grad = grad_fn(optimizer.target)
  optimizer = optimizer.apply_gradient(grad)
  return optimizer, {'loss': loss, 'accuracy': accuracy}


def train(data_path, master_csv_path, split_path, batch_size,
          num_training_steps, save_dir):
  """OGB Training Script."""

  # Initialize the dataset reader.
  reader = data_utils.DataReader(
      data_path=data_path,
      master_csv_path=master_csv_path,
      split_path=split_path,
      batch_size=batch_size)
  # Repeat the dataset forever for training.
  reader.repeat()

  net = GraphNetwork(mlp_features=(128, 128), latent_size=128)

  # Get a candidate graph and label to initialize the network.
  graph = reader.get_graph_by_idx(0)

  # Initialize the network.
  logging.info('Initializing network.')
  params = net.init(jax.random.PRNGKey(42), graph)
  optimizer = optim.Adam(learning_rate=1e-4).create(params)
  optimizer = jax.device_put(optimizer)

  for idx in range(num_training_steps):
    graph = next(reader)
    # Jax will re-jit your graphnet every time a new graph shape is encountered.
    # In the limit, this means a new compilation every training step, which
    # will result in *extremely* slow training. To prevent this, pad each
    # batch of graphs to the nearest power of two. Since jax maintains a cache
    # of compiled programs, the compilation cost is amortized.
    graph = pad_graph_to_nearest_power_of_two(graph)

    # Remove the label from the input graph/
    label = graph.globals['label']
    graph = graph._replace(globals={})

    optimizer, scalars = train_step(optimizer, graph, label, net)
    if idx % 100 == 0:
      logging.info('step: %s, loss: %s, acc: %s', idx, scalars['loss'],
                   scalars['accuracy'])
  if save_dir is not None:
    with pathlib.Path(save_dir, 'molhiv.pkl').open('wb') as fp:
      logging.info('Saving model to %s', save_dir)
      pickle.dump(optimizer.target, fp)
  logging.info('Training finished')


def evaluate(data_path, master_csv_path, split_path, save_dir):
  """Evaluation Script."""
  logging.info('Evaluating OGB molviv')
  logging.info('Dataset split: %s', split_path)
  # Initialize the dataset reader.
  reader = data_utils.DataReader(
      data_path=data_path,
      master_csv_path=master_csv_path,
      split_path=split_path,
      batch_size=1)

  with pathlib.Path(save_dir, 'molhiv.pkl').open('rb') as fp:
    params = pickle.load(fp)
  accumulated_loss = 0
  accumulated_accuracy = 0
  idx = 0

  # We jit the computation of our loss, since this is the main computation.
  # Using jax.jit means that we will use a single accelerator. If you want
  # to use more than 1 accelerator, use jax.pmap. More information can be
  # found in the jax documentation.
  net = GraphNetwork(mlp_features=[128, 128], latent_size=128)
  compute_loss_fn = jax.jit(functools.partial(compute_loss, net=net))
  for graph in reader:

    # Jax will re-jit your graphnet every time a new graph shape is encountered.
    # In the limit, this means a new compilation every training step, which
    # will result in *extremely* slow training. To prevent this, pad each
    # batch of graphs to the nearest power of two. Since jax maintains a cache
    # of compiled programs, the compilation cost is amortized.
    graph = pad_graph_to_nearest_power_of_two(graph)

    # Extract the labels and remove from the graph.
    label = graph.globals['label']
    graph = graph._replace(globals={})
    loss, acc = compute_loss_fn(params, graph, label)
    accumulated_accuracy += acc
    accumulated_loss += loss
    idx += 1
    if idx % 100 == 0:
      logging.info('Evaluated %s graphs', idx)
  logging.info('Completed evaluation.')
  loss = accumulated_loss / idx
  accuracy = accumulated_accuracy / idx
  logging.info('Eval loss: %s, accuracy %s', loss, accuracy)
  return loss, accuracy


def main(_):
  if FLAGS.mode == 'train':
    train(FLAGS.data_path, FLAGS.master_csv_path, FLAGS.split_path,
          FLAGS.batch_size, FLAGS.num_training_steps, FLAGS.save_dir)
  elif FLAGS.mode == 'evaluate':
    evaluate(FLAGS.data_path, FLAGS.master_csv_path, FLAGS.split_path,
             FLAGS.save_dir)

if __name__ == '__main__':
  app.run(main)



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A library of Graph Neural Network models."""

import functools
from typing import Any, Callable, Iterable, Mapping, Optional, Union

import jax
import jax.numpy as jnp
import jax.tree_util as tree
from jraph._src import graph as gn_graph
from jraph._src import utils

# As of 04/2020 pytype doesn't support recursive types.
# pytype: disable=not-supported-yet
ArrayTree = Union[jnp.ndarray, Iterable['ArrayTree'], Mapping[Any, 'ArrayTree']]

# All features will be an ArrayTree.
NodeFeatures = EdgeFeatures = SenderFeatures = ReceiverFeatures = Globals = ArrayTree

# Signature:
# (edges of each node to be aggregated, segment ids, number of segments) ->
# aggregated edges
AggregateEdgesToNodesFn = Callable[
    [EdgeFeatures, jnp.ndarray, int], NodeFeatures]

# Signature:
# (nodes of each graph to be aggregated, segment ids, number of segments) ->
# aggregated nodes
AggregateNodesToGlobalsFn = Callable[[NodeFeatures, jnp.ndarray, int],
                                     Globals]

# Signature:
# (edges of each graph to be aggregated, segment ids, number of segments) ->
# aggregated edges
AggregateEdgesToGlobalsFn = Callable[[EdgeFeatures, jnp.ndarray, int],
                                     Globals]

# Signature:
# (edge features, sender node features, receiver node features, globals) ->
# attention weights
AttentionLogitFn = Callable[
    [EdgeFeatures, SenderFeatures, ReceiverFeatures, Globals], ArrayTree]

# Signature:
# (edge features, weights) -> edge features for node update
AttentionReduceFn = Callable[[EdgeFeatures, ArrayTree], EdgeFeatures]

# Signature:
# (edges to be normalized, segment ids, number of segments) ->
# normalized edges
AttentionNormalizeFn = Callable[[EdgeFeatures, jnp.ndarray, int], EdgeFeatures]

# Signature:
# (edge features, sender node features, receiver node features, globals) ->
# updated edge features
GNUpdateEdgeFn = Callable[
    [EdgeFeatures, SenderFeatures, ReceiverFeatures, Globals], EdgeFeatures]

# Signature:
# (node features, outgoing edge features, incoming edge features,
#  globals) -> updated node features
GNUpdateNodeFn = Callable[
    [NodeFeatures, SenderFeatures, ReceiverFeatures, Globals], NodeFeatures]

GNUpdateGlobalFn = Callable[[NodeFeatures, EdgeFeatures, Globals], Globals]


def GraphNetwork(
    update_edge_fn: Optional[GNUpdateEdgeFn],
    update_node_fn: Optional[GNUpdateNodeFn],
    update_global_fn: Optional[GNUpdateGlobalFn] = None,
    aggregate_edges_for_nodes_fn: AggregateEdgesToNodesFn = utils.segment_sum,
    aggregate_nodes_for_globals_fn: AggregateNodesToGlobalsFn = utils
    .segment_sum,
    aggregate_edges_for_globals_fn: AggregateEdgesToGlobalsFn = utils
    .segment_sum,
    attention_logit_fn: Optional[AttentionLogitFn] = None,
    attention_normalize_fn: Optional[AttentionNormalizeFn] = utils
    .segment_softmax,
    attention_reduce_fn: Optional[AttentionReduceFn] = None):
  """Returns a method that applies a configured GraphNetwork.

  This implementation follows Algorithm 1 in https://arxiv.org/abs/1806.01261

  There is one difference. For the nodes update the class aggregates over the
  sender edges and receiver edges separately. This is a bit more general
  than the algorithm described in the paper. The original behaviour can be
  recovered by using only the receiver edge aggregations for the update.

  In addition this implementation supports softmax attention over incoming
  edge features.

  Example usage::

    gn = GraphNetwork(update_edge_function,
    update_node_function, **kwargs)
    # Conduct multiple rounds of message passing with the same parameters:
    for _ in range(num_message_passing_steps):
      graph = gn(graph)

  Args:
    update_edge_fn: function used to update the edges or None to deactivate edge
      updates.
    update_node_fn: function used to update the nodes or None to deactivate node
      updates.
    update_global_fn: function used to update the globals or None to deactivate
      globals updates.
    aggregate_edges_for_nodes_fn: function used to aggregate messages to each
      node.
    aggregate_nodes_for_globals_fn: function used to aggregate the nodes for the
      globals.
    aggregate_edges_for_globals_fn: function used to aggregate the edges for the
      globals.
    attention_logit_fn: function used to calculate the attention weights or
      None to deactivate attention mechanism.
    attention_normalize_fn: function used to normalize raw attention logits or
      None if attention mechanism is not active.
    attention_reduce_fn: function used to apply weights to the edge features or
      None if attention mechanism is not active.

  Returns:
    A method that applies the configured GraphNetwork.
  """
  not_both_supplied = lambda x, y: (x != y) and ((x is None) or (y is None))
  if not_both_supplied(attention_reduce_fn, attention_logit_fn):
    raise ValueError(('attention_logit_fn and attention_reduce_fn must both be'
                      ' supplied.'))

  def _ApplyGraphNet(graph):
    """Applies a configured GraphNetwork to a graph.

    This implementation follows Algorithm 1 in https://arxiv.org/abs/1806.01261

    There is one difference. For the nodes update the class aggregates over the
    sender edges and receiver edges separately. This is a bit more general
    the algorithm described in the paper. The original behaviour can be
    recovered by using only the receiver edge aggregations for the update.

    In addition this implementation supports softmax attention over incoming
    edge features.

    Many popular Graph Neural Networks can be implemented as special cases of
    GraphNets, for more information please see the paper.

    Args:
      graph: a `GraphsTuple` containing the graph.

    Returns:
      Updated `GraphsTuple`.
    """
    # pylint: disable=g-long-lambda
    nodes, edges, receivers, senders, globals_, n_node, n_edge = graph
    # Equivalent to jnp.sum(n_node), but jittable
    sum_n_node = tree.tree_leaves(nodes)[0].shape[0]
    sum_n_edge = senders.shape[0]
    if not tree.tree_all(
        tree.tree_map(lambda n: n.shape[0] == sum_n_node, nodes)):
      raise ValueError(
          'All node arrays in nest must contain the same number of nodes.')

    sent_attributes = tree.tree_map(lambda n: n[senders], nodes)
    received_attributes = tree.tree_map(lambda n: n[receivers], nodes)
    # Here we scatter the global features to the corresponding edges,
    # giving us tensors of shape [num_edges, global_feat].
    global_edge_attributes = tree.tree_map(lambda g: jnp.repeat(
        g, n_edge, axis=0, total_repeat_length=sum_n_edge), globals_)

    if update_edge_fn:
      edges = update_edge_fn(edges, sent_attributes, received_attributes,
                             global_edge_attributes)

    if attention_logit_fn:
      logits = attention_logit_fn(edges, sent_attributes, received_attributes,
                                  global_edge_attributes)
      tree_calculate_weights = functools.partial(
          attention_normalize_fn,
          segment_ids=receivers,
          num_segments=sum_n_node)
      weights = tree.tree_map(tree_calculate_weights, logits)
      edges = attention_reduce_fn(edges, weights)

    if update_node_fn:
      sent_attributes = tree.tree_map(
          lambda e: aggregate_edges_for_nodes_fn(e, senders, sum_n_node), edges)
      received_attributes = tree.tree_map(
          lambda e: aggregate_edges_for_nodes_fn(e, receivers, sum_n_node),
          edges)
      # Here we scatter the global features to the corresponding nodes,
      # giving us tensors of shape [num_nodes, global_feat].
      global_attributes = tree.tree_map(lambda g: jnp.repeat(
          g, n_node, axis=0, total_repeat_length=sum_n_node), globals_)
      nodes = update_node_fn(nodes, sent_attributes,
                             received_attributes, global_attributes)

    if update_global_fn:
      n_graph = n_node.shape[0]
      graph_idx = jnp.arange(n_graph)
      # To aggregate nodes and edges from each graph to global features,
      # we first construct tensors that map the node to the corresponding graph.
      # For example, if you have `n_node=[1,2]`, we construct the tensor
      # [0, 1, 1]. We then do the same for edges.
      node_gr_idx = jnp.repeat(
          graph_idx, n_node, axis=0, total_repeat_length=sum_n_node)
      edge_gr_idx = jnp.repeat(
          graph_idx, n_edge, axis=0, total_repeat_length=sum_n_edge)
      # We use the aggregation function to pool the nodes/edges per graph.
      node_attributes = tree.tree_map(
          lambda n: aggregate_nodes_for_globals_fn(n, node_gr_idx, n_graph),
          nodes)
      edge_attribtutes = tree.tree_map(
          lambda e: aggregate_edges_for_globals_fn(e, edge_gr_idx, n_graph),
          edges)
      # These pooled nodes are the inputs to the global update fn.
      globals_ = update_global_fn(node_attributes, edge_attribtutes, globals_)
    # pylint: enable=g-long-lambda
    return gn_graph.GraphsTuple(
        nodes=nodes,
        edges=edges,
        receivers=receivers,
        senders=senders,
        globals=globals_,
        n_node=n_node,
        n_edge=n_edge)

  return _ApplyGraphNet


InteractionUpdateNodeFn = Callable[
    [NodeFeatures,
     Mapping[str, SenderFeatures],
     Mapping[str, ReceiverFeatures]],
    NodeFeatures]
InteractionUpdateNodeFnNoSentEdges = Callable[
    [NodeFeatures,
     Mapping[str, ReceiverFeatures]],
    NodeFeatures]

InteractionUpdateEdgeFn = Callable[
    [EdgeFeatures, SenderFeatures, ReceiverFeatures], EdgeFeatures]


def InteractionNetwork(
    update_edge_fn: InteractionUpdateEdgeFn,
    update_node_fn: Union[InteractionUpdateNodeFn,
                          InteractionUpdateNodeFnNoSentEdges],
    aggregate_edges_for_nodes_fn: AggregateEdgesToNodesFn = utils.segment_sum,
    include_sent_messages_in_node_update: bool = False):
  """Returns a method that applies a configured InteractionNetwork.

  An interaction network computes interactions on the edges based on the
  previous edges features, and on the features of the nodes sending into those
  edges. It then updates the nodes based on the incoming updated edges.
  See https://arxiv.org/abs/1612.00222 for more details.

  This implementation adds an option not in https://arxiv.org/abs/1612.00222,
  which is to include edge features for which a node is a sender in the
  arguments to the node update function.

  Args:
    update_edge_fn: a function mapping a single edge update inputs to a single
      edge feature.
    update_node_fn: a function mapping a single node update input to a single
      node feature.
    aggregate_edges_for_nodes_fn: function used to aggregate messages to each
      node.
    include_sent_messages_in_node_update: pass edge features for which a node is
      a sender to the node update function.
  """
  # An InteractionNetwork is a GraphNetwork without globals features,
  # so we implement the InteractionNetwork as a configured GraphNetwork.

  # An InteractionNetwork edge function does not have global feature inputs,
  # so we filter the passed global argument in the GraphNetwork.
  wrapped_update_edge_fn = lambda e, s, r, g: update_edge_fn(e, s, r)

  # Similarly, we wrap the update_node_fn to ensure only the expected
  # arguments are passed to the Interaction net.
  if include_sent_messages_in_node_update:
    wrapped_update_node_fn = lambda n, s, r, g: update_node_fn(n, s, r)
  else:
    wrapped_update_node_fn = lambda n, s, r, g: update_node_fn(n, r)
  return GraphNetwork(
      update_edge_fn=wrapped_update_edge_fn,
      update_node_fn=wrapped_update_node_fn,
      aggregate_edges_for_nodes_fn=aggregate_edges_for_nodes_fn)


# Signature:
# edge features -> embedded edge features
EmbedEdgeFn = Callable[[EdgeFeatures], EdgeFeatures]

# Signature:
# node features -> embedded node features
EmbedNodeFn = Callable[[NodeFeatures], NodeFeatures]

# Signature:
# globals features -> embedded globals features
EmbedGlobalFn = Callable[[Globals], Globals]


def GraphMapFeatures(embed_edge_fn: Optional[EmbedEdgeFn] = None,
                     embed_node_fn: Optional[EmbedNodeFn] = None,
                     embed_global_fn: Optional[EmbedGlobalFn] = None):
  """Returns function which embeds the components of a graph independently.

  Args:
    embed_edge_fn: function used to embed the edges.
    embed_node_fn: function used to embed the nodes.
    embed_global_fn: function used to embed the globals.
  """
  identity = lambda x: x
  embed_edges_fn = embed_edge_fn if embed_edge_fn else identity
  embed_nodes_fn = embed_node_fn if embed_node_fn else identity
  embed_global_fn = embed_global_fn if embed_global_fn else identity

  def Embed(graphs_tuple):
    return graphs_tuple._replace(
        nodes=embed_nodes_fn(graphs_tuple.nodes),
        edges=embed_edges_fn(graphs_tuple.edges),
        globals=embed_global_fn(graphs_tuple.globals))

  return Embed


def RelationNetwork(
    update_edge_fn: Callable[[SenderFeatures, ReceiverFeatures], EdgeFeatures],
    update_global_fn: Callable[[EdgeFeatures], NodeFeatures],
    aggregate_edges_for_globals_fn:
        AggregateEdgesToGlobalsFn = utils.segment_sum):
  """Returns a method that applies a Relation Network.

  See https://arxiv.org/abs/1706.01427 for more details.

  This implementation has one more argument, `aggregate_edges_for_globals_fn`,
  which changes how edge features are aggregated. The paper uses the default -
  `utils.segment_sum`.

  Args:
    update_edge_fn: function used to update the edges.
    update_global_fn: function used to update the globals.
    aggregate_edges_for_globals_fn: function used to aggregate the edges for the
      globals.
  """
  return GraphNetwork(
      update_edge_fn=lambda e, s, r, g: update_edge_fn(s, r),
      update_node_fn=None,
      update_global_fn=lambda n, e, g: update_global_fn(e),
      attention_logit_fn=None,
      aggregate_edges_for_globals_fn=aggregate_edges_for_globals_fn)


def DeepSets(
    update_node_fn: Callable[[NodeFeatures, Globals], NodeFeatures],
    update_global_fn: Callable[[NodeFeatures], Globals],
    aggregate_nodes_for_globals_fn:
        AggregateNodesToGlobalsFn = utils.segment_sum):
  """Returns a method that applies a DeepSets layer.

  Implementation for the model described in https://arxiv.org/abs/1703.06114
  (M. Zaheer, S. Kottur, S. Ravanbakhsh, B. Poczos, R. Salakhutdinov, A. Smola).
  See also PointNet (https://arxiv.org/abs/1612.00593, C. Qi, H. Su, K. Mo,
  L. J. Guibas) for a related model.

  This module operates on sets, which can be thought of as graphs without
  edges. The nodes features are first updated based on their value and the
  globals features, and new globals features are then computed based on the
  updated nodes features.

  Args:
    update_node_fn: function used to update the nodes.
    update_global_fn: function used to update the globals.
    aggregate_nodes_for_globals_fn: function used to aggregate the nodes for the
      globals.
  """
  # DeepSets can be implemented with a GraphNetwork, with just a node
  # update function that takes nodes and globals, and a global update
  # function based on the updated node features.
  return GraphNetwork(
      update_edge_fn=None,
      update_node_fn=lambda n, s, r, g: update_node_fn(n, g),
      update_global_fn=lambda n, e, g: update_global_fn(n),
      aggregate_nodes_for_globals_fn=aggregate_nodes_for_globals_fn)


def GraphNetGAT(
    update_edge_fn: GNUpdateEdgeFn,
    update_node_fn: GNUpdateNodeFn,
    attention_logit_fn: AttentionLogitFn,
    attention_reduce_fn: AttentionReduceFn,
    update_global_fn: Optional[GNUpdateGlobalFn] = None,
    aggregate_edges_for_nodes_fn: AggregateEdgesToNodesFn = utils.segment_sum,
    aggregate_nodes_for_globals_fn: AggregateNodesToGlobalsFn = utils.
    segment_sum,
    aggregate_edges_for_globals_fn: AggregateEdgesToGlobalsFn = utils.
    segment_sum
    ):
  """Returns a method that applies a GraphNet with attention on edge features.

  Args:
    update_edge_fn: function used to update the edges.
    update_node_fn: function used to update the nodes.
    attention_logit_fn: function used to calculate the attention weights.
    attention_reduce_fn: function used to apply attention weights to the edge
      features.
    update_global_fn: function used to update the globals or None to deactivate
      globals updates.
    aggregate_edges_for_nodes_fn: function used to aggregate attention-weighted
      messages to each node.
    aggregate_nodes_for_globals_fn: function used to aggregate the nodes for the
      globals.
    aggregate_edges_for_globals_fn: function used to aggregate
      attention-weighted edges for the globals.

  Returns:
    A function that applies a GraphNet Graph Attention layer.
  """
  if (attention_logit_fn is None) or (attention_reduce_fn is None):
    raise ValueError(('`None` value not supported for `attention_logit_fn` or '
                      '`attention_reduce_fn` in a Graph Attention network.'))
  return GraphNetwork(
      update_edge_fn=update_edge_fn,
      update_node_fn=update_node_fn,
      update_global_fn=update_global_fn,
      attention_logit_fn=attention_logit_fn,
      attention_reduce_fn=attention_reduce_fn,
      aggregate_edges_for_nodes_fn=aggregate_edges_for_nodes_fn,
      aggregate_nodes_for_globals_fn=aggregate_nodes_for_globals_fn,
      aggregate_edges_for_globals_fn=aggregate_edges_for_globals_fn)


GATAttentionQueryFn = Callable[[NodeFeatures], NodeFeatures]
GATAttentionLogitFn = Callable[
    [SenderFeatures, ReceiverFeatures, EdgeFeatures], EdgeFeatures]
GATNodeUpdateFn = Callable[[NodeFeatures], NodeFeatures]


def GAT(attention_query_fn: GATAttentionQueryFn,
        attention_logit_fn: GATAttentionLogitFn,
        node_update_fn: Optional[GATNodeUpdateFn] = None):
  """Returns a method that applies a Graph Attention Network layer.

  Graph Attention message passing as described in
  https://arxiv.org/abs/1710.10903. This model expects node features as a
  jnp.array, may use edge features for computing attention weights, and
  ignore global features. It does not support nests.

  NOTE: this implementation assumes that the input graph has self edges. To
  recover the behavior of the referenced paper, please add self edges.

  Args:
    attention_query_fn: function that generates attention queries
      from sender node features.
    attention_logit_fn: function that converts attention queries into logits for
      softmax attention.
    node_update_fn: function that updates the aggregated messages. If None,
      will apply leaky relu and concatenate (if using multi-head attention).

  Returns:
    A function that applies a Graph Attention layer.
  """
  # pylint: disable=g-long-lambda
  if node_update_fn is None:
    # By default, apply the leaky relu and then concatenate the heads on the
    # feature axis.
    node_update_fn = lambda x: jnp.reshape(
        jax.nn.leaky_relu(x), (x.shape[0], -1))
  def _ApplyGAT(graph):
    """Applies a Graph Attention layer."""
    nodes, edges, receivers, senders, _, _, _ = graph
    # Equivalent to the sum of n_node, but statically known.
    try:
      sum_n_node = nodes.shape[0]
    except IndexError:
      raise IndexError('GAT requires node features')  # pylint: disable=raise-missing-from

    # First pass nodes through the node updater.
    nodes = attention_query_fn(nodes)
    # pylint: disable=g-long-lambda
    # We compute the softmax logits using a function that takes the
    # embedded sender and receiver attributes.
    sent_attributes = nodes[senders]
    received_attributes = nodes[receivers]
    softmax_logits = attention_logit_fn(
        sent_attributes, received_attributes, edges)

    # Compute the softmax weights on the entire tree.
    weights = utils.segment_softmax(softmax_logits, segment_ids=receivers,
                                    num_segments=sum_n_node)
    # Apply weights
    messages = sent_attributes * weights
    # Aggregate messages to nodes.
    nodes = utils.segment_sum(messages, receivers, num_segments=sum_n_node)

    # Apply an update function to the aggregated messages.
    nodes = node_update_fn(nodes)
    return graph._replace(nodes=nodes)
  # pylint: enable=g-long-lambda
  return _ApplyGAT


def GraphConvolution(
    update_node_fn: Callable[[NodeFeatures], NodeFeatures],
    aggregate_nodes_fn: AggregateEdgesToNodesFn = utils.segment_sum,
    add_self_edges: bool = False,
    symmetric_normalization: bool = True):
  """Returns a method that applies a Graph Convolution layer.

  Graph Convolutional layer as in https://arxiv.org/abs/1609.02907,

  NOTE: This implementation does not add an activation after aggregation.
  If you are stacking layers, you may want to add an activation between
  each layer.

  Args:
    update_node_fn: function used to update the nodes. In the paper a single
      layer MLP is used.
    aggregate_nodes_fn: function used to aggregates the sender nodes.
    add_self_edges: whether to add self edges to nodes in the graph as in the
      paper definition of GCN. Defaults to False.
    symmetric_normalization: whether to use symmetric normalization. Defaults
      to True. Note that to replicate the fomula of the linked paper, the
      adjacency matrix must be symmetric. If the adjacency matrix is not
      symmetric the data is prenormalised by the sender degree matrix and post
      normalised by the receiver degree matrix.

  Returns:
    A method that applies a Graph Convolution layer.
  """
  def _ApplyGCN(graph):
    """Applies a Graph Convolution layer."""
    nodes, _, receivers, senders, _, _, _ = graph

    # First pass nodes through the node updater.
    nodes = update_node_fn(nodes)
    # Equivalent to jnp.sum(n_node), but jittable
    total_num_nodes = tree.tree_leaves(nodes)[0].shape[0]
    if add_self_edges:
      # We add self edges to the senders and receivers so that each node
      # includes itself in aggregation.
      # In principle, a `GraphsTuple` should partition by n_edge, but in
      # this case it is not required since a GCN is agnostic to whether
      # the `GraphsTuple` is a batch of graphs or a single large graph.
      conv_receivers = jnp.concatenate((receivers, jnp.arange(total_num_nodes)),
                                       axis=0)
      conv_senders = jnp.concatenate((senders, jnp.arange(total_num_nodes)),
                                     axis=0)
    else:
      conv_senders = senders
      conv_receivers = receivers

    # pylint: disable=g-long-lambda
    if symmetric_normalization:
      # Calculate the normalization values.
      count_edges = lambda x: utils.segment_sum(
          jnp.ones_like(conv_senders), x, total_num_nodes)
      sender_degree = count_edges(conv_senders)
      receiver_degree = count_edges(conv_receivers)

      # Pre normalize by sqrt sender degree.
      # Avoid dividing by 0 by taking maximum of (degree, 1).
      nodes = tree.tree_map(
          lambda x: x * jax.lax.rsqrt(jnp.maximum(sender_degree, 1.0))[:, None],
          nodes,
      )
      # Aggregate the pre normalized nodes.
      nodes = tree.tree_map(
          lambda x: aggregate_nodes_fn(x[conv_senders], conv_receivers,
                                       total_num_nodes), nodes)
      # Post normalize by sqrt receiver degree.
      # Avoid dividing by 0 by taking maximum of (degree, 1).
      nodes = tree.tree_map(
          lambda x:
          (x * jax.lax.rsqrt(jnp.maximum(receiver_degree, 1.0))[:, None]),
          nodes,
      )
    else:
      nodes = tree.tree_map(
          lambda x: aggregate_nodes_fn(x[conv_senders], conv_receivers,
                                       total_num_nodes), nodes)
    # pylint: enable=g-long-lambda
    return graph._replace(nodes=nodes)

  return _ApplyGCN



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for jraph.ogb_examples.train_pmap."""

import pathlib
from absl.testing import absltest
from jraph.ogb_examples import train_pmap


class TrainTest(absltest.TestCase):

  def test_train_and_eval_overfit(self):
    ogb_path = pathlib.Path(train_pmap.__file__).parents[0]
    master_csv_path = pathlib.Path(ogb_path, 'test_data', 'master.csv')
    split_path = pathlib.Path(ogb_path, 'test_data', 'train.csv.gz')
    data_path = master_csv_path.parents[0]
    temp_dir = self.create_tempdir().full_path
    train_pmap.train(data_path, master_csv_path, split_path, 1, 101, temp_dir)
    _, accuracy = train_pmap.evaluate(data_path, master_csv_path, split_path,
                                      temp_dir)
    self.assertEqual(float(accuracy), 1.0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of Conway's game of life using jraph."""

import time

from absl import app
import haiku as hk
import jax
import jax.numpy as jnp
import jraph
import numpy as np


def conway_mlp(x):
  """Implements a MLP representing Conway's game of life rules."""
  w = jnp.array([[0.0, -1.0], [0.0, 1.0], [0.0, 1.0],
                 [0, -1.0], [1.0, 1.0], [1.0, 1.0]])
  b = jnp.array([3.5, -3.5, -1.5, 1.5, -2.5, -3.5])
  h = jnp.maximum(jnp.dot(w, x) + b, 0.)
  w = jnp.array([[2.0, -4.0, 2.0, -4.0, 2.0, -4.0]])
  b = jnp.array([-4.0])
  y = jnp.maximum(jnp.dot(w, h) + b, 0.0)
  return y


def conway_graph(size) -> jraph.GraphsTuple:
  """Returns a graph representing the game field of conway's game of life."""
  # Creates nodes: each node represents a cell in the game.
  n_node = size**2
  nodes = np.zeros((n_node, 1))
  node_indices = jnp.arange(n_node)
  # Creates edges, senders and receivers:
  # the senders represent the connections to the 8 neighboring fields.
  n_edge = 8 * n_node
  edges = jnp.zeros((n_edge, 1))
  senders = jnp.vstack(
      [node_indices - size - 1, node_indices - size, node_indices - size + 1,
       node_indices - 1, node_indices + 1,
       node_indices + size - 1, node_indices + size, node_indices + size + 1])
  senders = senders.T.reshape(-1)
  senders = (senders + size**2) % size**2
  receivers = jnp.repeat(node_indices, 8)
  # Adds a glider to the game
  nodes[0, 0] = 1.0
  nodes[1, 0] = 1.0
  nodes[2, 0] = 1.0
  nodes[2 + size, 0] = 1.0
  nodes[1 + 2 * size, 0] = 1.0
  return jraph.GraphsTuple(n_node=jnp.array([n_node]),
                           n_edge=jnp.array([n_edge]),
                           nodes=jnp.asarray(nodes),
                           edges=edges,
                           globals=None,
                           senders=senders,
                           receivers=receivers)


def display_graph(graph: jraph.GraphsTuple):
  """Prints the nodes of the graph representing Conway's game of life."""
  size = int(np.sqrt(np.sum(graph.n_node)))

  def _display_node(node):
    if node == 1.0:
      return 'x'
    else:
      return ' '

  nodes = graph.nodes.copy()
  output = '\n'.join(
      ''.join(_display_node(nodes[i * size + j][0])
              for j in range(size))
      for i in range(size))
  print('-' * size + '\n' + output)


def main(_):

  def net_fn(graph: jraph.GraphsTuple):
    unf = jraph.concatenated_args(conway_mlp)
    net = jraph.InteractionNetwork(
        update_edge_fn=lambda e, n_s, n_r: n_s,
        update_node_fn=jax.vmap(unf))
    return net(graph)

  net = hk.without_apply_rng(hk.transform(net_fn))

  cg = conway_graph(size=20)
  params = net.init(jax.random.PRNGKey(42), cg)
  for _ in range(100):
    time.sleep(0.05)
    cg = jax.jit(net.apply)(params, cg)
    display_graph(cg)

if __name__ == '__main__':
  app.run(main)



# Copyright 2020 DeepMind Technologies Limited.


# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Example of a Hamiltonian Graph Network (https://arxiv.org/abs/1909.12790).

In this example:
  * The `GraphNetwork`s implements the hardcoded formulas of the Hooke's
    Hamiltonian to return the scalar hamiltonian of a system
    (see `hookes_hamiltonian_from_graph_fn`).
  * Then JAX autodiff is used to obtain the function of derivatives of the state
    via Hamilton equations(see `get_state_derivatives_from_hamiltonian_fn`).
  * The function of the derivatives of the state is used by a generic integrator
    to simulate steps forward in time (see `single_integration_step`).
  * `build_hookes_particle_state_graph` is used to sample the initial state
    of the system.

Note this example does not implement a learned Hamiltonian Graph Network, but
a hardcoded Hamiltonian function corresponding to the Hooke's potential,
implemented as a `GraphNetwork`. This is to show how natural is to express the
true Hamiltonian of a particle system as a Graph Network. However, the only
system-specific code here is the hard coded Hamiltonian in
`hookes_hamiltonian_from_graph_fn` and the data generation provided in
`build_hookes_particle_state_graph`. Everything is reusable for any other
system.

To implement a learned Hamiltonian Graph Network, one could closely follow
`hookes_hamiltonian_from_graph_fn` using function approximators (e.g. MLP) in
`edge_update_fn`, `node_update_fn` and `global_update_fn` that take as inputs
the concatenated features (e.g. using `jraph.concatenated_args`), with the only
condition that the `global_update_fn` has an output size of 1, to match
the expected output size for a Hamiltonian. Then the learned Hamiltonian Graph
Network would be trained by simply adding a loss term in the position / momentum
of the next step after applying the integrator.


Note it is recommended to use immutable container types to store nested edge,
node and global features to avoid unwanted side effects. In this example we
use `frozendict`s, which we register with `jax.tree_util`.

"""

import functools
from typing import Tuple, Callable

from absl import app
from frozendict import frozendict
import jax
import jax.numpy as jnp
import jraph
import matplotlib.pyplot as plt
import numpy as np


# Tell tree_util how to navigate frozendicts.
jax.tree_util.register_pytree_node(
    frozendict,
    flatten_func=lambda s: (tuple(s.values()), tuple(s.keys())),
    unflatten_func=lambda k, xs: frozendict(zip(k, xs)))


def hookes_hamiltonian_from_graph_fn(
    graph: jraph.GraphsTuple) -> jraph.GraphsTuple:
  """Computes Hamiltonian of a Hooke's potential system represented in a graph.

  While this function hardcodes the Hamiltonian for a Hooke's potential, a
  learned Hamiltonian Graph Network (https://arxiv.org/abs/1909.12790) could
  be implemented by replacing the hardcoded formulas by learnable MLPs that
  take as inputs all of the concatenated features to the edge_fn, node_fn,
  and global_fn, and outputs a single scalar value in the global_fn.

  Args:
    graph: `GraphsTuple` where the nodes contain:
        - "mass": [num_particles]
        - "position": [num_particles, num_dims]
        - "momentum": [num_particles, num_dims]
        and the edges contain:
        - "spring_constant": [num_interations]

  Returns:
    `GraphsTuple` with features:
        - edge features: "hookes_potential" [num_interactions]
        - node features: "kinetic_energy" [num_particles]
        - global features: "hamiltonian" [batch_size]

  """

  def update_edge_fn(edges, senders, receivers, globals_):
    del globals_
    distance = jnp.linalg.norm(senders["position"] - receivers["position"])
    hookes_potential_per_edge = 0.5 * edges["spring_constant"] * distance ** 2
    return frozendict({"hookes_potential": hookes_potential_per_edge})

  def update_node_fn(nodes, sent_edges, received_edges, globals_):
    del sent_edges, received_edges, globals_
    momentum_norm = jnp.linalg.norm(nodes["momentum"])
    kinetic_energy_per_node = momentum_norm ** 2 / (2 * nodes["mass"])
    return frozendict({"kinetic_energy": kinetic_energy_per_node})

  def update_global_fn(nodes, edges, globals_):
    del globals_
    # At this point we will receive node and edge features aggregated (summed)
    # for all nodes and edges in each graph.
    hamiltonian_per_graph = nodes["kinetic_energy"] + edges["hookes_potential"]
    return frozendict({"hamiltonian": hamiltonian_per_graph})

  gn = jraph.GraphNetwork(
      update_edge_fn=update_edge_fn,
      update_node_fn=update_node_fn,
      update_global_fn=update_global_fn)

  return gn(graph)


# Methods for generating the data.
def build_hookes_particle_state_graph(num_particles: int) -> jraph.GraphsTuple:
  """Generates a graph representing a Hooke's system in a random state."""

  mass = np.random.uniform(0, 5, [num_particles])
  velocity = get_random_uniform_norm2d_vectors(0, 0.1, num_particles)
  position = get_random_uniform_norm2d_vectors(0, 1, num_particles)
  momentum = velocity * np.expand_dims(mass, axis=-1)
  # Remove average momentum, so center of mass does not move.
  momentum = momentum - momentum.mean(0, keepdims=True)

  # Connect all particles to all particles.
  particle_indices = np.arange(num_particles)
  senders, receivers = np.meshgrid(particle_indices, particle_indices)
  senders, receivers = senders.flatten(), receivers.flatten()

  # Generate a symmetric random matrix of spring constants.
  # Generate random elements stringly in the lower triangular part.
  spring_constants = np.random.uniform(
      1e-2, 1e-1, [num_particles, num_particles])
  spring_constants = np.tril(
      spring_constants) + np.tril(spring_constants, -1).T
  spring_constants = spring_constants.flatten()

  # Remove interactions of particles to themselves.
  mask = senders != receivers
  senders, receivers = senders[mask], receivers[mask]
  spring_constants = spring_constants[mask]
  num_interactions = receivers.shape[0]

  return jraph.GraphsTuple(
      n_node=np.asarray([num_particles]),
      n_edge=np.asarray([num_interactions]),
      nodes={
          "mass": mass,  # Scalar mass for each particle.
          "position": position,  # 2d position for each particle.
          "momentum": momentum,  # 2d momentum for each particle.
      },
      edges={
          # Scalar spring constant for each interaction
          "spring_constant": spring_constants,
      },
      globals={},
      senders=senders,
      receivers=receivers)


def get_random_uniform_norm2d_vectors(
    min_norm: float, max_norm: float, num_particles: int) -> np.ndarray:
  """Returns 2-d vectors with random norms."""
  norm = np.random.uniform(min_norm, max_norm, [num_particles, 1])
  angle = np.random.uniform(0, 2*np.pi, [num_particles])
  return norm * np.stack([np.cos(angle), np.sin(angle)], axis=-1)


def get_fully_connected_senders_and_receivers(
    num_particles: int, self_edges: bool = False,
    ) -> Tuple[np.ndarray, np.ndarray]:
  """Returns senders and receivers for fully connected particles."""
  particle_indices = np.arange(num_particles)
  senders, receivers = np.meshgrid(particle_indices, particle_indices)
  senders, receivers = senders.flatten(), receivers.flatten()
  if not self_edges:
    mask = senders != receivers
    senders, receivers = senders[mask], receivers[mask]
  return senders, receivers


# All code below here is general purpose for any system or integrator.
# Utility methods for getting/setting the state of the particles in the graph
# (position and momentum), and for obtaining the static part of the graph
# (connectivity and particle parameters: masses, spring constants).
def set_system_state(
    static_graph: jraph.GraphsTuple,
    position: np.ndarray,
    momentum: np.ndarray) -> jraph.GraphsTuple:
  """Sets the non-static parameters of the graph (momentum, position)."""
  nodes = static_graph.nodes.copy(position=position, momentum=momentum)
  return static_graph._replace(nodes=nodes)


def get_system_state(graph: jraph.GraphsTuple) -> Tuple[np.ndarray, np.ndarray]:
  return graph.nodes["position"], graph.nodes["momentum"]


def get_static_graph(graph: jraph.GraphsTuple) -> jraph.GraphsTuple:
  """Returns the graph with the static parts of a system only."""
  nodes = dict(graph.nodes)
  del nodes["position"], nodes["momentum"]
  return graph._replace(nodes=frozendict(nodes))


# Utility methods to operate with Hamiltonian functions.
def get_hamiltonian_from_state_fn(
    static_graph: jraph.GraphsTuple,
    hamiltonian_from_graph_fn: Callable[[jraph.GraphsTuple], jraph.GraphsTuple],
    ) -> Callable[[np.ndarray, np.ndarray], float]:
  """Returns fn such that fn(position, momentum) -> scalar Hamiltonian.

  Args:
      static_graph: `GraphsTuple` containing per-particle static parameters and
         connectivity, such as a full graph of the state can be build by calling
         `set_system_state(static_graph, position, momentum)`.
      hamiltonian_from_graph_fn: callable that given an input `GraphsTuple`
         returns a `GraphsTuple` with a "hamiltonian" field in the globals.

  Returns:
      Function that given a state (position, momentum) returns the scalar
      Hamiltonian.
  """

  def hamiltonian_from_state_fn(position, momentum):
    # Note we sum along the batch dimension to get the total energy in the batch
    # so get can easily get the gradient.
    graph = set_system_state(static_graph, position, momentum)
    output_graph = hamiltonian_from_graph_fn(graph)
    return output_graph.globals["hamiltonian"].sum()

  return hamiltonian_from_state_fn


def get_state_derivatives_from_hamiltonian_fn(
    hamiltonian_from_state_fn: Callable[[np.ndarray, np.ndarray], float],
    ) -> Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
  """Returns fn(position, momentum, ...) -> (dposition_dt, dmomentum_dt).

  Args:
      hamiltonian_from_state_fn: Function that given a state
          (position, momentum)  returns the scalar Hamiltonian.

  Returns:
      Function that given a state (position, momentum) returns the time
      derivatives of the state (dposition_dt, dmomentum_dt) by applying
      Hamilton equations.

  """

  hamiltonian_gradients_fn = jax.grad(hamiltonian_from_state_fn, argnums=[0, 1])

  def state_derivatives_from_hamiltonian_fn(
      position: np.ndarray, momentum: np.ndarray
      ) -> Tuple[np.ndarray, np.ndarray]:
    # Take the derivatives against position and momentum.
    dh_dposition, dh_dmomentum = hamiltonian_gradients_fn(position, momentum)

    # Hamilton equations.
    dposition_dt = dh_dmomentum
    dmomentum_dt = - dh_dposition
    return dposition_dt, dmomentum_dt
  return state_derivatives_from_hamiltonian_fn


# Implementations of some general purpose integrators for Hamiltonian states.
StateDerivativesFnType = Callable[
    [np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]


def abstract_integrator(
    position: np.ndarray, momentum: np.ndarray, time_step: float,
    state_derivatives_fn: StateDerivativesFnType,
    ) -> Tuple[np.ndarray, np.ndarray]:
  """Signature of an abstract integrator.

  An integrator is a function, that given the the current state, a time step,
  and a `state_derivatives_fn` returns the next state.

  Args:
      position: array with the position at time t.
      momentum: array with the momentum at time t.
      time_step: integration step size.
      state_derivatives_fn: a function fn, that returns time derivatives of a
          state such fn(position, momentum) -> (dposition_dt, dmomentum_dt)
          where dposition_dt, dmomentum_dt, have the same shapes as
          position, momentum.

  Returns:
      Tuple with position and momentum at time `t + time_step`.

  """
  raise NotImplementedError("Abstract integrator")


def euler_integrator(
    position: np.ndarray, momentum: np.ndarray, time_step: float,
    state_derivatives_fn: StateDerivativesFnType,
    ) -> Tuple[np.ndarray, np.ndarray]:
  """Implementation of an Euler integrator (see `abstract_integrator`)."""
  dposition_dt, dmomentum_dt = state_derivatives_fn(position, momentum)
  next_position = position + dposition_dt * time_step
  next_momentum = momentum + dmomentum_dt * time_step
  return next_position, next_momentum


def verlet_integrator(
    position: np.ndarray, momentum: np.ndarray, time_step: float,
    state_derivatives_fn: StateDerivativesFnType,
    ) -> Tuple[np.ndarray, np.ndarray]:
  """Implementation of Verlet integrator (see `abstract_integrator`)."""

  _, dmomentum_dt = state_derivatives_fn(position, momentum)
  aux_momentum = momentum + dmomentum_dt * time_step / 2

  dposition_dt, _ = state_derivatives_fn(position, aux_momentum)
  next_position = position + dposition_dt * time_step

  _, dmomentum_dt = state_derivatives_fn(next_position, aux_momentum)
  next_momentum = aux_momentum + dmomentum_dt * time_step / 2

  return next_position, next_momentum


# Single graph -> graph integration step.
IntegratorType = Callable[
    [np.ndarray, np.ndarray, float, StateDerivativesFnType],
    Tuple[np.ndarray, np.ndarray]
]


def single_integration_step(
    graph: jraph.GraphsTuple, time_step: float,
    integrator_fn: IntegratorType,
    hamiltonian_from_graph_fn: Callable[[jraph.GraphsTuple], jraph.GraphsTuple],
    ) -> Tuple[float, jraph.GraphsTuple]:
  """Updates a graph state integrating by a single step.

  Args:
    graph: `GraphsTuple` representing a system state at time t.
    time_step: size of the timestep to integrate for.
    integrator_fn: Integrator to use. A function fn such that
       fn(position_t, momentum_t, time_step, state_derivatives_fn) ->
           (position_tp1, momentum_tp1)
    hamiltonian_from_graph_fn: Function that given a `GraphsTuple`, returns
        another one with a "hamiltonian" global field.

  Returns:
    `GraphsTuple` representing a system state at time `t + time_step`.

  """

  # Template graph with particle/interactions parameters and connectiviity
  # but without the state (position/momentum).
  static_graph = get_static_graph(graph)

  # Get the Hamiltonian function, and the function that returns the state
  # derivatives.
  hamiltonian_fn = get_hamiltonian_from_state_fn(
      static_graph=static_graph,
      hamiltonian_from_graph_fn=hamiltonian_from_graph_fn)
  state_derivatives_fn = get_state_derivatives_from_hamiltonian_fn(
      hamiltonian_fn)

  # Get the current state.
  position, momentum = get_system_state(graph)

  # Calling the integrator to get the next state.
  next_position, next_momentum = integrator_fn(
      position, momentum, time_step, state_derivatives_fn)
  next_graph = set_system_state(static_graph, next_position, next_momentum)

  # Return the energy of the next state too for plotting.
  energy = hamiltonian_fn(next_position, next_momentum)

  return energy, next_graph


def main(_):

  # Get a state function and jit it.
  # We could switch to any other Hamiltonian and any other integrator here.
  # e.g. the non-symplectic `euler_integrator`.
  step_fn = functools.partial(
      single_integration_step,
      hamiltonian_from_graph_fn=hookes_hamiltonian_from_graph_fn,
      integrator_fn=verlet_integrator)
  step_fn = jax.jit(step_fn)

  # Get a graph with the initial state.
  num_particles = 10
  graph = build_hookes_particle_state_graph(num_particles)

  # Iterate for multiple timesteps.
  num_steps = 200
  time_step = 0.002
  positions_sequence = []
  total_energies = []
  steps = []
  for step_i in range(num_steps):
    energy, graph = step_fn(graph, time_step)
    total_energies.append(energy)
    positions_sequence.append(graph.nodes["position"])
    steps.append(step_i + 1)

  # Plot results (positions and energy as a function of time).
  unused_fig, axes = plt.subplots(1, 2, figsize=(15, 5))
  positions_sequence_array = np.stack(positions_sequence, axis=0)
  axes[0].plot(positions_sequence_array[..., 0],
               positions_sequence_array[..., -1])
  axes[0].set_xlabel("Particle position x")
  axes[0].set_ylabel("Particle position y")

  axes[1].plot(steps, total_energies)
  axes[1].set_ylim(0, max(total_energies)*1.2)
  axes[1].set_xlabel("Simulation step")
  axes[1].set_ylabel("Total energy")
  plt.show()

if __name__ == "__main__":
  app.run(main)
